import socket
import user_agents
from hiddifypanel.cache import cache
from sqlalchemy.orm import Load
import glob
import json
from babel.dates import format_timedelta as babel_format_timedelta
from flask_babelex import gettext as __
from flask_babelex import lazy_gettext as _
from hiddifypanel.models import *
from hiddifypanel.panel.database import db
import datetime
from flask import jsonify, g, url_for, Markup, abort, current_app, request
from flask import flash as flask_flash
import re
from wtforms.validators import ValidationError
import requests

import string
import random
from babel.dates import format_timedelta as babel_format_timedelta
import urllib
import time
import os
import psutil
from urllib.parse import urlparse
import ssl
import h2.connection
import subprocess
import netifaces
import time

to_gig_d = 1000*1000*1000


def add_temporary_access():
    random_port = random.randint(30000, 50000)
    exec_command(
        f'sudo /opt/hiddify-config/hiddify-panel/temporary_access.sh {random_port} &')
    temp_admin_link = f"http://{get_ip(4)}:{random_port}{get_admin_path()}"
    g.temp_admin_link = temp_admin_link


def add_short_link(link, period_min=5):
    # pattern = "\^/([^/]+)(/)?\?\$\ {return 302 " + re.escape(link) + ";}"
    pattern = r"([^/]+)\("

    with open(current_app.config['HIDDIFY_CONFIG_PATH']+"/nginx/parts/short-link.conf", 'r') as f:
        for line in f:
            if link in line:
                return re.search(pattern, line).group(1)

    short_code = get_random_string(6, 10).lower()
    exec_command(
        f'sudo /opt/hiddify-config/nginx/add2shortlink.sh {link} {short_code} {period_min} &')
    return short_code


def get_admin_path():
    proxy_path = hconfig(ConfigEnum.proxy_path)
    admin_secret = g.admin.uuid or get_super_admin_secret()
    return (f"/{proxy_path}/{admin_secret}/admin/")


def exec_command(cmd, cwd=None):
    try:
        subprocess.Popen(cmd.split(" "))  # run in background
    except Exception as e:
        print(e)


def auth(function):
    def wrapper(*args, **kwargs):
        if g.user_uuid == None:
            return jsonify({"error": "auth failed"})
        if not admin and g.is_admin:
            return jsonify({"error": "admin can not access user page. add /admin/ to your url"})
        return function()

    return wrapper


def super_admin(function):
    def wrapper(*args, **kwargs):
        if g.admin.mode not in [AdminMode.super_admin]:
            abort(403, __("Access Denied"))
            return jsonify({"error": "auth failed"})
        return function(*args, **kwargs)

    return wrapper


def admin(function):
    def wrapper(*args, **kwargs):
        if g.admin.mode not in [AdminMode.admin, AdminMode.super_admin]:
            abort(_("Access Denied"), 403)
            return jsonify({"error": "auth failed"})

        return function(*args, **kwargs)
    return wrapper


def abs_url(path):
    return f"/{g.proxy_path}/{g.user_uuid}/{path}"


def asset_url(path):
    return f"/{g.proxy_path}/{path}"


@cache.cache(ttl=600)
def get_direct_host_or_ip(prefer_version):
    direct = Domain.query.filter(Domain.mode == DomainType.direct, Domain.sub_link_only == False).first()
    if not (direct):
        direct = Domain.query.filter(Domain.mode == DomainType.direct).first()
    if direct:
        direct = direct.domain
    else:
        direct = get_ip(prefer_version)
    if not direct:
        direct = get_ip(4 if prefer_version == 6 else 6)
    return direct


@cache.cache(ttl=600)
def get_ip(version, retry=5):
    ips = get_interface_public_ip(version)
    ip = None
    if (ips):
        ip = random.sample(ips, 1)[0]

    if ip is None:
        ip = get_socket_public_ip(version)

    if ip is None:
        try:
            ip = urllib.request.urlopen(f'https://v{version}.ident.me/').read().decode('utf8')

        except:
            pass
    if ip is None and retry > 0:
        ip = get_ip(version, retry=retry-1)
    return ip


def get_socket_public_ip(version):
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        if version == 6:
            s.connect(("2001:4860:4860::8888", 80))
        else:
            s.connect(("8.8.8.8", 80))
        ip_address = s.getsockname()[0]
        s.close()

        return ip_address if is_public_ip(ip_address) else None
    except socket.error:
        return None


def is_public_ip(address):
    if address.startswith('127.') or address.startswith('169.254.') or address.startswith('10.') or address.startswith('192.168.') or address.startswith('172.'):
        return False
    if address.startswith('fe80:') or address.startswith('fd') or address.startswith('fc00:'):
        return False
    return True


def get_interface_public_ip(version):
    addresses = []
    try:
        interfaces = netifaces.interfaces()
        for interface in interfaces:
            if version == 4:
                address_info = netifaces.ifaddresses(interface).get(netifaces.AF_INET, [])
            elif version == 6:
                address_info = netifaces.ifaddresses(interface).get(netifaces.AF_INET6, [])
            else:
                return None

            if address_info:
                for addr in address_info:
                    address = addr['addr']
                    if (is_public_ip(address)):
                        addresses.append(address)

        return addresses if addresses else None

    except (OSError, KeyError):
        return None


@cache.cache(ttl=300)
def get_available_proxies(child_id):
    proxies = Proxy.query.filter(Proxy.child_id == child_id).all()
    proxies = [c for c in proxies if 'restls' not in c.transport]
    if not hconfig(ConfigEnum.domain_fronting_domain, child_id):
        proxies = [c for c in proxies if 'Fake' not in c.cdn]
    if not hconfig(ConfigEnum.ssfaketls_enable, child_id):
        proxies = [c for c in proxies if 'faketls' != c.transport]
    if not hconfig(ConfigEnum.v2ray_enable, child_id):
        proxies = [c for c in proxies if 'v2ray' != c.proto]
    if not hconfig(ConfigEnum.shadowtls_enable, child_id):
        proxies = [c for c in proxies if c.transport != 'shadowtls']
    if not hconfig(ConfigEnum.ssr_enable, child_id):
        proxies = [c for c in proxies if 'ssr' != c.proto]
    if not hconfig(ConfigEnum.vmess_enable, child_id):
        proxies = [c for c in proxies if 'vmess' not in c.proto]

    if not hconfig(ConfigEnum.kcp_enable, child_id):
        proxies = [c for c in proxies if 'kcp' not in c.l3]

    if not hconfig(ConfigEnum.http_proxy_enable, child_id):
        proxies = [c for c in proxies if 'http' != c.l3]

    if not Domain.query.filter(Domain.mode.in_([DomainType.cdn, DomainType.auto_cdn_ip])).first():
        proxies = [c for c in proxies if c.cdn != "CDN"]

    proxies = [c for c in proxies if not ('vless' == c.proto and ProxyTransport.tcp == c.transport and c.cdn == ProxyCDN.direct)]
    return proxies


def quick_apply_users():
    if hconfig(ConfigEnum.is_parent):
        return
    # from hiddifypanel.panel import usage
    # usage.update_local_usage()
    # return
    # for user in User.query.all():
    #     if is_user_active(user):
    #         xray_api.add_client(user.uuid)
    #     else:
    #         xray_api.remove_client(user.uuid)

    exec_command("sudo /opt/hiddify-config/install.sh apply_users &")

    time.sleep(1)
    return {"status": 'success'}


def flash_config_success(restart_mode='', domain_changed=True):
    if restart_mode:
        url = url_for('admin.Actions:reinstall', complete_install=restart_mode ==
                      'reinstall', domain_changed=domain_changed)
        apply_btn = f"<a href='{url}' class='btn btn-primary form_post'>" + \
            _("admin.config.apply_configs")+"</a>"
        flash((_('config.validation-success', link=apply_btn)), 'success')
    else:
        flash((_('config.validation-success-no-reset')), 'success')


# Importing socket library

# Function to display hostname and
# IP address


def get_domain_ip(dom, retry=3, version=None):

    res = None
    if not version:
        try:
            res = socket.gethostbyname(dom)
        except:
            pass

    if not res and version != 6:
        try:
            res = socket.getaddrinfo(dom, None, socket.AF_INET)[0][4][0]
        except:
            pass

    if not res and version != 4:
        try:
            res = f"[{socket.getaddrinfo(dom, None, socket.AF_INET6)[0][4][0]}]"
        except:
            pass

    if retry <= 0:
        return None

    return res or get_domain_ip(dom, retry=retry-1)


def check_connection_to_remote(api_url):

    path = f"{api_url}/api/v1/hello/"

    try:
        res = requests.get(path, verify=False, timeout=2).json()
        return True

    except:
        return False


def check_connection_for_domain(domain):

    proxy_path = hconfig(ConfigEnum.proxy_path)
    admin_secret = hconfig(ConfigEnum.admin_secret)
    path = f"{proxy_path}/{admin_secret}/api/v1/hello/"
    try:
        print(f"https://{domain}/{path}")
        res = requests.get(
            f"https://{domain}/{path}", verify=False, timeout=10).json()
        return res['status'] == 200

    except:
        try:
            print(f"http://{domain}/{path}")
            res = requests.get(
                f"http://{domain}/{path}", verify=False, timeout=10).json()
            return res['status'] == 200
        except:
            try:
                print(f"http://{get_domain_ip(domain)}/{path}")
                res = requests.get(
                    f"http://{get_domain_ip(domain)}/{path}", verify=False, timeout=10).json()
                return res['status'] == 200
            except:
                return False
    return True


def get_user_link(uuid, domain, mode='', username=''):
    is_cdn = domain.mode == DomainType.cdn if type(domain) == Domain else False
    proxy_path = hconfig(ConfigEnum.proxy_path)
    res = ""
    if mode == "multi":
        res += "<div class='btn-group'>"
    d = domain.domain
    if "*" in d:
        d = d.replace("*", get_random_string(5, 15))

    link = f"https://{d}/{proxy_path}/{uuid}/#{username}"
    if mode == "admin":
        link = f"https://{d}/{proxy_path}/{uuid}/admin/#{username}"
    link_multi = f"{link}multi"
    # if mode == 'new':
    #     link = f"{link}new"
    text = domain.alias or domain.domain
    color_cls = 'info'

    if type(domain) == Domain and not domain.sub_link_only and domain.mode in [DomainType.cdn, DomainType.auto_cdn_ip]:
        auto_cdn = (domain.mode == DomainType.auto_cdn_ip) or (domain.cdn_ip and "MTN" in domain.cdn_ip)
        color_cls = "success" if auto_cdn else 'warning'
        text = f'<span class="badge badge-secondary" >{"Auto" if auto_cdn else "CDN"}</span> '+text

    if mode == "multi":
        res += f"<a class='btn btn-xs btn-secondary' target='_blank' href='{link_multi}' >{_('all')}</a>"
    res += f"<a target='_blank' href='{link}' class='btn btn-xs btn-{color_cls} ltr' ><i class='fa-solid fa-arrow-up-right-from-square d-none'></i> {text}</a>"

    if mode == "multi":
        res += "</div>"
    return res


def flash(message, category):
    print(message)
    return flask_flash(Markup(message), category)


def validate_domain_exist(form, field):
    domain = field.data
    if not domain:
        return
    dip = get_domain_ip(domain)
    if dip == None:
        raise ValidationError(
            _("Domain can not be resolved! there is a problem in your domain"))


def reinstall_action(complete_install=False, domain_change=False, do_update=False):
    from hiddifypanel.panel.admin.Actions import Actions
    action = Actions()
    if do_update:
        return action.update()
    return action.reinstall(complete_install=complete_install, domain_changed=domain_changed)


def check_need_reset(old_configs, do=False):
    restart_mode = ''
    for c in old_configs:
        # c=ConfigEnum(c)
        if old_configs[c] != hconfig(c) and c.apply_mode():
            if restart_mode != 'reinstall':
                restart_mode = c.apply_mode()

    # do_full_install=old_config[ConfigEnum.telegram_lib]!=hconfig(ConfigEnum.telegram_lib)
    if old_configs[ConfigEnum.package_mode] != hconfig(ConfigEnum.package_mode):
        return reinstall_action(do_update=True)
    if not (do and restart_mode == 'reinstall'):
        return flash_config_success(restart_mode=restart_mode, domain_changed=False)

    return reinstall_action(complete_install=True, domain_changed=domain_changed)


def format_timedelta(delta, add_direction=True, granularity="days"):
    res = delta.days
    locale = g.locale if g and hasattr(g, "locale") else hconfig(ConfigEnum.admin_lang)
    if granularity == "days" and delta.days == 0:
        res = _("0 - Last day")
    elif delta.days < 7 or delta.days >= 60:
        res = babel_format_timedelta(delta, threshold=1, add_direction=add_direction, locale=locale)
    elif delta.days < 60:
        res = babel_format_timedelta(delta, granularity="day", threshold=10, add_direction=add_direction, locale=locale)
    return res


def get_child(unique_id):
    child_id = 0
    if unique_id is None:
        child_id = 0
    else:
        child = Child.query.filter(Child.unique_id == str(unique_id)).first()
        if not child:
            child = Child(unique_id=str(unique_id))
            db.session.add(child)
            db.session.commit()
            child = Child.query.filter(Child.unique_id == str(unique_id)).first()
        child_id = child.id
    return child_id


def date_to_json(d):
    return d.strftime("%Y-%m-%d") if d else None


def json_to_date(date_str):
    try:
        return datetime.datetime.strptime(date_str, '%Y-%m-%d')
    except:
        return None


def time_to_json(d):
    return d.strftime("%Y-%m-%d %H:%M:%S") if d else None


def json_to_time(time_str):
    try:
        return datetime.datetime.strptime(time_str, "%Y-%m-%d %H:%M:%S")
    except:
        return None


def dump_db_to_dict():
    return {"users": [u.to_dict() for u in User.query.all()],
            "domains": [u.to_dict() for u in Domain.query.all()],
            "proxies": [u.to_dict() for u in Proxy.query.all()],
            "parent_domains": [] if not hconfig(ConfigEnum.license) else [u.to_dict() for u in ParentDomain.query.all()],
            'admin_users': [d.to_dict() for d in AdminUser.query.all()],
            "hconfigs": [*[u.to_dict() for u in BoolConfig.query.all()],
                         *[u.to_dict() for u in StrConfig.query.all()]]
            }


def set_db_from_json(json_data, override_child_id=None, set_users=True, set_domains=True, set_proxies=True, set_settings=True, remove_domains=False, remove_users=False, override_unique_id=True, set_admins=True):
    new_rows = []
    if set_users and 'users' in json_data:
        bulk_register_users(json_data['users'], commit=False, remove=remove_users)
    if set_admins and 'admin_users' in json_data:
        bulk_register_admins(json_data['admin_users'], commit=False)
    if set_domains and 'domains' in json_data:
        bulk_register_domains(json_data['domains'], commit=False, remove=remove_domains, override_child_id=override_child_id)
    if set_domains and 'parent_domains' in json_data:
        bulk_register_parent_domains(json_data['parent_domains'], commit=False, remove=remove_domains)
    if set_settings and 'hconfigs' in json_data:
        bulk_register_configs(json_data["hconfigs"], commit=False, override_child_id=override_child_id, override_unique_id=override_unique_id)
        if 'proxies' in json_data:
            bulk_register_proxies(json_data['proxies'], commit=False, override_child_id=override_child_id)
    db.session.commit()


def get_random_string(min_=10, max_=30):
    # With combination of lower and upper case
    length = random.randint(min_, max_)
    characters = string.ascii_letters + string.digits
    result_str = ''.join(random.choice(characters) for i in range(length))
    return result_str


def get_warp_info():
    proxies = dict(http='socks5://127.0.0.1:3000',
                   https='socks5://127.0.0.1:3000')
    res = requests.get("https://cloudflare.com/cdn-cgi/trace", proxies=proxies, timeout=1).text

    dicres = {line.split("=")[0]: line.split("=")[0] for line in res}
    return dicres


def system_stats():
    # CPU usage
    cpu_percent = psutil.cpu_percent(interval=1)

    # RAM usage
    ram_stats = psutil.virtual_memory()
    ram_used = ram_stats.used / 1024**3
    ram_total = ram_stats.total / 1024**3

    # Disk usage (in GB)
    disk_stats = psutil.disk_usage('/')
    disk_used = disk_stats.used / 1024**3
    disk_total = disk_stats.total / 1024**3

    hiddify_used = get_folder_size('/opt/hiddify-config/') / 1024**3

    # Network usage
    net_stats = psutil.net_io_counters()
    bytes_sent_cumulative = net_stats.bytes_sent
    bytes_recv_cumulative = net_stats.bytes_recv
    bytes_sent = net_stats.bytes_sent - getattr(system_stats, 'prev_bytes_sent', 0)
    bytes_recv = net_stats.bytes_recv - getattr(system_stats, 'prev_bytes_recv', 0)
    system_stats.prev_bytes_sent = net_stats.bytes_sent
    system_stats.prev_bytes_recv = net_stats.bytes_recv

    # Total connections and unique IPs
    connections = psutil.net_connections()
    total_connections = len(connections)
    unique_ips = set([conn.raddr.ip for conn in connections if conn.status == 'ESTABLISHED' and conn.raddr])
    total_unique_ips = len(unique_ips)

    # Load average
    num_cpus = psutil.cpu_count()
    load_avg = [avg / num_cpus for avg in os.getloadavg()]
    # Return the system information
    return {
        "cpu_percent": cpu_percent / num_cpus,
        "ram_used": ram_used,
        "ram_total": ram_total,
        "disk_used": disk_used,
        "disk_total": disk_total,
        "hiddify_used": hiddify_used,
        "bytes_sent": bytes_sent,
        "bytes_recv": bytes_recv,
        "bytes_sent_cumulative": bytes_sent_cumulative,
        "bytes_recv_cumulative": bytes_recv_cumulative,
        "net_sent_cumulative_GB": bytes_sent_cumulative / 1024**3,
        "net_total_cumulative_GB": (bytes_sent_cumulative+bytes_recv_cumulative) / 1024**3,
        "total_connections": total_connections,
        "total_unique_ips": total_unique_ips,
        "load_avg_1min": load_avg[0],
        "load_avg_5min": load_avg[1],
        "load_avg_15min": load_avg[2],
        'num_cpus': num_cpus
    }


def top_processes():
    # Get the process information
    processes = [p for p in psutil.process_iter(['name', 'memory_full_info', 'cpu_percent']) if p.info['name'] != '']
    num_cores = psutil.cpu_count()
    # Calculate memory usage, RAM usage, and CPU usage for each process
    memory_usage = {}
    ram_usage = {}
    cpu_usage = {}
    for p in processes:
        name = p.info['name']
        if "python3" in name or "uwsgi" in name or 'flask' in name:
            name = "Hiddify"
        mem_info = p.info['memory_full_info']
        if mem_info is None:
            continue
        mem_usage = mem_info.uss
        cpu_percent = p.info['cpu_percent']/num_cores
        if name in memory_usage:
            memory_usage[name] += mem_usage / (1024 ** 3)
            ram_usage[name] += mem_info.rss / (1024 ** 3)
            cpu_usage[name] += cpu_percent
        else:
            memory_usage[name] = mem_usage / (1024 ** 3)
            ram_usage[name] = mem_info.rss / (1024 ** 3)
            cpu_usage[name] = cpu_percent

    while len(cpu_usage) < 5:
        cpu_usage[" "*len(cpu_usage)] = 0
    while len(ram_usage) < 5:
        ram_usage[" "*len(ram_usage)] = 0
    while len(memory_usage) < 5:
        memory_usage[" "*len(memory_usage)] = 0
    # Sort the processes by memory usage, RAM usage, and CPU usage
    top_memory = sorted(memory_usage.items(), key=lambda x: x[1], reverse=True)[:5]
    top_ram = sorted(ram_usage.items(), key=lambda x: x[1], reverse=True)[:5]
    top_cpu = sorted(cpu_usage.items(), key=lambda x: x[1], reverse=True)[:5]

    # Return the top processes for memory usage, RAM usage, and CPU usage
    return {
        "memory": top_memory,
        "ram": top_ram,
        "cpu": top_cpu
    }


def get_folder_size(folder_path):
    total_size = 0
    try:
        for dirpath, dirnames, filenames in os.walk(folder_path):
            for file in filenames:
                file_path = os.path.join(dirpath, file)
                try:
                    total_size += os.path.getsize(file_path)
                except:
                    pass
    except:
        pass
    return total_size


def get_domain_btn_link(domain):
    text = domain.alias or domain.domain
    color_cls = "info"
    if domain.mode in [DomainType.cdn, DomainType.auto_cdn_ip]:
        auto_cdn = (domain.mode == DomainType.auto_cdn_ip) or (domain.cdn_ip and "MTN" in domain.cdn_ip)
        color_cls = "success" if auto_cdn else 'warning'
        text = f'<span class="badge badge-secondary" >{"Auto" if auto_cdn else "CDN"}</span> '+text
    res = f"<a target='_blank' href='#' class='btn btn-xs btn-{color_cls} ltr' ><i class='fa-solid fa-arrow-up-right-from-square d-none'></i> {text}</a>"
    return res


def get_random_domains(count=1, retry=3):
    try:
        irurl = "https://api.ooni.io/api/v1/measurements?probe_cc=IR&test_name=web_connectivity&anomaly=false&confirmed=false&failure=false&order_by=test_start_time&limit=1000"
        # cnurl="https://api.ooni.io/api/v1/measurements?probe_cc=CN&test_name=web_connectivity&anomaly=false&confirmed=false&failure=false&order_by=test_start_time&limit=1000"
        data_ir = requests.get(irurl).json()
        # data_cn=requests.get(url).json()

        domains = [urlparse(d['input']).netloc.lower() for d in data_ir['results'] if d['scores']['blocking_country'] == 0.0]
        domains = [d for d in domains if not d.endswith(".ir") and not ".gov" in d]

        return random.sample(domains, count)
    except Exception as e:
        print('Error, getting random domains... ', e, 'retrying...', retry)
        if retry <= 0:
            defdomains = ["fa.wikipedia.org", 'en.wikipedia.org', 'wikipedia.org', 'yahoo.com', 'en.yahoo.com']
            print('Error, using default domains')
            return random.sample(defdomains, count)
        return get_random_domains(count, retry-1)


def is_domain_support_tls_13(domain):
    context = ssl.create_default_context()
    with socket.create_connection((domain, port)) as sock:
        with context.wrap_socket(sock, server_hostname=domain) as ssock:
            return ssock.version() == "TLSv1.3"


def is_domain_support_h2(sni, server=None):
    try:

        context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
        context.options |= (ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 | ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1)
        context.options |= ssl.OP_NO_COMPRESSION
        context.set_ciphers("ECDHE+AESGCM:ECDHE+CHACHA20:DHE+AESGCM:DHE+CHACHA20")
        context.set_alpn_protocols(["h2"])
        start_time = time.monotonic()
        with socket.create_connection((server or sni, 443), timeout=2) as sock:

            with context.wrap_socket(sock, server_hostname=sni) as ssock:
                elapsed_time = time.monotonic() - start_time
                valid = ssock.version() == "TLSv1.3"
                if valid:
                    return int(max(1, elapsed_time*1000))
                return False
    except Exception as e:
        print(f'{domain} {e}')
        return False


def is_domain_reality_friendly(domain):
    return is_domain_support_h2(domain)


def debug_flash_if_not_in_the_same_asn(domain):
    from hiddifypanel.panel.clean_ip import ipasn
    ipv4 = get_ip(4)
    dip = get_domain_ip(domain)
    try:
        if ipasn:
            asn_ipv4 = ipasn.get(ipv4)
            asn_dip = ipasn.get(dip)
            # country_ipv4= ipcountry.get(ipv4)
            # country_dip= ipcountry.get(dip)
            if asn_ipv4.get('autonomous_system_organization') != asn_dip.get('autonomous_system_organization'):
                flash(_("selected domain for REALITY is not in the same ASN. To better use of the protocol, it is better to find a domain in the same ASN.") +
                      f"<br> Server ASN={asn_ipv4.get('autonomous_system_organization','unknown')}<br>{domain}_ASN={asn_dip.get('autonomous_system_organization','unknown')}", "warning")
    except:
        pass


def fallback_domain_compatible_with_servernames(fallback_domain, servername):
    return is_domain_support_h2(servername, fallback_domain)


def generate_x25519_keys():
    # Run the "xray x25519" command and capture its output
    try:
        cmd = "xray x25519"
        output = subprocess.check_output(cmd, shell=True, text=True)
    except:
        output = """
        Private key: ILnwK6Ii9PWgnkq5Lbb8G_chyP4ba5cGRpZbaJjf7lg
        Public key: no36JbbL8uPH4VT2PNe9husJBN2mu5DbgDASH7hK32A
        """
        # Extract the private and public keys from the output
    private_key = output.split("Private key: ")[1].split("\n")[0]
    public_key = output.split("Public key: ")[1].split("\n")[0]

    # Return the keys as a tuple
    return {"private_key": private_key, "public_key": public_key}


def get_hostkeys(dojson=False):
    key_files = glob.glob(current_app.config['HIDDIFY_CONFIG_PATH'] + "/other/ssh/host_key/*_key.pub")
    host_keys = []
    for file_name in key_files:
        with open(file_name, "r") as f:
            host_key = f.read().strip()
            host_key = host_key.split()
            if len(host_key) > 2:
                host_key = host_key[:2]  # strip the hostname part
            host_key = " ".join(host_key)
            host_keys.append(host_key)
    if dojson:
        return json.dumps(host_keys)
    return host_keys


def get_ssh_client_version(user):
    return 'SSH-2.0-OpenSSH_7.4p1'


def get_ed25519_private_public_pair():
    from cryptography.hazmat.primitives.asymmetric import ed25519
    from cryptography.hazmat.primitives import serialization
    privkey = ed25519.Ed25519PrivateKey.generate()
    pubkey = privkey.public_key()
    priv_bytes = privkey.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.OpenSSH,
        encryption_algorithm=serialization.NoEncryption(),
    )
    pub_bytes = pubkey.public_bytes(
        encoding=serialization.Encoding.OpenSSH,
        format=serialization.PublicFormat.OpenSSH,
    )
    return priv_bytes.decode(), pub_bytes.decode()


def do_base_64(str):
    import base64
    resp = base64.b64encode(f'{str}'.encode("utf-8"))
    return resp.decode()


def get_user_agent():
    return __parse_user_agent(request.user_agent.string)


@cache.cache()
def __parse_user_agent(ua):
    uaa = user_agents.parse(request.user_agent.string)
    res = {}
    res["is_browser"] = re.match('^Mozilla', ua, re.IGNORECASE) and True
    res['os'] = uaa.os.family
    res['os_version'] = uaa.os.version
    res['is_clash'] = re.match('^(Clash|Stash)', ua, re.IGNORECASE) and True
    res['is_clash_meta'] = re.match('^(Clash-verge|Clash-?Meta|Stash|NekoBox|NekoRay|Pharos|hiddify-desktop)', ua, re.IGNORECASE) and True
    res['is_singbox'] = re.match('^(HiddifyNext|Dart|SFI|SFA)', ua, re.IGNORECASE) and True
    if (res['is_singbox']):
        res['singbox_version'] = (1, 4, 0)
    res['is_hiddify'] = re.match('^(HiddifyNext)', ua, re.IGNORECASE) and True
    if ['is_hiddify']:
        res['hiddify_version'] = uaa
    res['is_v2ray'] = re.match('^(Hiddify|FoXray|Fair|v2rayNG|SagerNet|Shadowrocket|V2Box|Loon|Liberty)', ua, re.IGNORECASE) and True

    if res['os'] == 'Other':
        if re.match('^(FoXray|Fair|Shadowrocket|V2Box|Loon|Liberty)', ua, re.IGNORECASE):
            res['os'] = 'iOS'
            # res['os_version']

    for a in ['Hiddify', 'FoXray', 'Fair', 'v2rayNG', 'SagerNet', 'Shadowrocket', 'V2Box', 'Loon', 'Liberty', 'Clash', 'Meta', 'Stash', 'SFI', 'SFA', 'HiddifyNext']:
        if a.lower() in ua.lower():
            res['app'] = a
    if res["is_browser"]:
        res['app'] = uaa.browser.family
    return res


def url_encode(strr):
    import urllib.parse
    return urllib.parse.quote(strr)


def error(str):
    import sys
    print(str, file=sys.stderr)


def static_url_for(**values):
    orig = url_for("static", **values)
    return orig.split("user_secret")[0]


def get_latest_release_version(repo_name):
    try:
        url = f"https://github.com/hiddify/{repo_name}/releases/latest"
        response = requests.head(url, allow_redirects=False)

        location_header = response.headers.get("Location")
        if location_header:
            version = re.search(r"/([^/]+)/?$", location_header)
            if version:
                return version.group(1).replace('v', '')
    except Exception as e:
        return f'{e}'

    return None



from flask_admin.contrib import sqla

from hiddifypanel.panel.database import db
from wtforms.validators import Regexp
from hiddifypanel.models import *
from wtforms.validators import Regexp, ValidationError
from .adminlte import AdminLTEModelView
from flask_babelex import gettext as __
from flask_babelex import lazy_gettext as _
from hiddifypanel.panel import hiddify,cf_api
from flask import Markup
from flask import Flask,g,url_for
from flask_sqlalchemy import SQLAlchemy
from flask_admin import Admin
from flask_admin.contrib.sqla import ModelView
from wtforms import SelectMultipleField
import datetime

from wtforms.widgets import ListWidget, CheckboxInput
from sqlalchemy.orm import backref
# Define a custom field type for the related domains
from flask_admin.form.fields import Select2TagsField,Select2Field
from wtforms import SelectField

class AdminModeField(SelectField):
    def __init__(self, label=None, validators=None, **kwargs):
        super(AdminModeField, self).__init__(label, validators, **kwargs)
        if g.admin.mode in [AdminMode.agent,AdminMode.admin]:
            self.choices = [ (AdminMode.agent.value, 'agent')]
        elif g.admin.mode==AdminMode.admin:
            self.choices = [ (AdminMode.agent.value, 'agent'),(AdminMode.admin.value, 'Admin'),]
        elif g.admin.mode==AdminMode.super_admin:
            self.choices = [(AdminMode.agent.value, 'agent'),(AdminMode.admin.value, 'Admin'),(AdminMode.super_admin.value, 'Super Admin')]



class SubAdminsField(SelectField):
    def __init__(self, label=None, validators=None,*args, **kwargs):
        kwargs.pop("allow_blank")
        super().__init__(label, validators,*args, **kwargs)
        self.choices=[(admin.id,admin.name) for admin in g.admin.sub_admins]
        self.choices+=[(g.admin.id,g.admin.name)]
class AdminstratorAdmin(AdminLTEModelView):
    column_hide_backrefs = False
    column_list = ["name",'UserLinks','mode','can_add_admin','max_active_users','max_users','online_users','comment']
    form_columns = ["name",'mode','can_add_admin','max_active_users','max_users','comment',"uuid"]
    list_template = 'model/admin_list.html'
    # edit_modal = True
    # form_overrides = {'work_with': Select2Field}
    
    form_overrides = {
        'mode': AdminModeField,
        'parent_admin': SubAdminsField
    }
    column_labels = {
        "Actions":_("actions"),
        "UserLinks":_("user.user_links"),
        "name": _("user.name"),
        "mode":_("Mode"),
        "uuid":_("user.UUID"),
        "comment":_("Note"),
        'max_active_users':_("Max Active Users"),
        'max_users':_('Max Users'),
        "online_users":_("Online Users"),
        'can_add_admin':_("Can add sub admin")

    }
    form_args = {
    'uuid': {
        'validators': [Regexp(r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$',message=__("Should be a valid uuid"))]
    #     'label': 'First Name',
    #     'validators': [required()]
    }}
    
    column_descriptions = dict(
        comment=_("Add some text that is only visible to super_admin."),
        mode=_("Define the admin mode. "),
    )
    # create_modal = True
    can_export = False
    
    # column_list = ["domain",'sub_link_only', "mode","alias", "domain_ip", "cdn_ip"]
    # column_editable_list=["domain"]
    # column_filters=["domain","mode"]
  
    
    column_searchable_list = ["name", "uuid"]
    
    # form_columns=['domain','sub_link_only','alias','mode','cdn_ip','show_domains']

    def _ul_formatter(view, context, model, name):
        
        return Markup(" ".join([hiddify.get_user_link(model.uuid,d,'admin',model.name) for d in get_panel_domains()]))
    
    @property
    def can_create(self):
        return g.admin.can_add_admin or g.admin.mode==AdminMode.super_admin
    def _name_formatter(view, context, model, name):
        proxy_path=hconfig(ConfigEnum.proxy_path)
        d=get_panel_domains()[0]
        if d:
            link=f"<a target='_blank' href='/{proxy_path}/{model.uuid}/admin/#{model.name}'>{model.name} <i class='fa-solid fa-arrow-up-right-from-square'></i></a>"
            if model.parent_admin:
                return Markup(model.parent_admin.name +"&rlm;&lrm; / &rlm;&lrm;"+link)
            return Markup(link)
        else:
            return model.name

    def _online_users_formatter(view, context, model, name):
        last_day=datetime.datetime.now()-datetime.timedelta(days=1)
        u=model.recursive_users_query().filter(User.last_online>last_day).count()
        t=model.recursive_users_query().count()
        # actives=[u for u in model.recursive_users_query().all() if is_user_active(u)]
        # allusers=model.recursive_users_query().count()
        # onlines=[p for p in  users  if p.last_online and p.last_online>last_day]
        # return Markup(f"<a class='btn btn-xs btn-default' href='{url_for('flask.user.index_view',admin_id=model.id)}'> {_('Online')}: {onlines}</a>")
        rate=round(u*100/(t+0.000001))
        state= "danger" if u>=t else ('warning' if  rate>80 else 'success')
        color= "#ff7e7e" if u>=t else ('#ffc107' if  rate>80 else '#9ee150')
        return Markup(f"""
        <div class="progress progress-lg position-relative" style="min-width: 100px;">
          <div class="progress-bar progress-bar-striped" role="progressbar" style="width: {rate}%;background-color: {color};" aria-valuenow="{rate}" aria-valuemin="0" aria-valuemax="100"></div>
              <span class='badge position-absolute' style="left:auto;right:auto;width: 100%;font-size:1em">{u} {_('user.home.usage.from')} {t}</span>

        </div>
        """)
    def _max_users_formatter(view, context, model, name):
        u=model.recursive_users_query().count()
        if model.mode== AdminMode.super_admin:
            return f"{u} / ∞"
        t=model.max_users
        rate=round(u*100/(t+0.000001))
        state= "danger" if u>=t else ('warning' if  rate>80 else 'success')
        color= "#ff7e7e" if u>=t else ('#ffc107' if  rate>80 else '#9ee150')
        return Markup(f"""
        <div class="progress progress-lg position-relative" style="min-width: 100px;">
          <div class="progress-bar progress-bar-striped" role="progressbar" style="width: {rate}%;background-color: {color};" aria-valuenow="{rate}" aria-valuemin="0" aria-valuemax="100"></div>
              <span class='badge position-absolute' style="left:auto;right:auto;width: 100%;font-size:1em">{u} {_('user.home.usage.from')} {t}</span>

        </div>
        """)
        
    def _max_active_users_formatter(view, context, model, name):
        
        actives=[u for u in model.recursive_users_query().all() if is_user_active(u)]
        u=len(actives)
        if model.mode== AdminMode.super_admin:
            return f"{u} / ∞"
        t=model.max_active_users
        rate=round(u*100/(t+0.000001))
        state= "danger" if u>=t else ('warning' if  rate>80 else 'success')
        color= "#ff7e7e" if u>=t else ('#ffc107' if  rate>80 else '#9ee150')
        return Markup(f"""
        <div class="progress progress-lg position-relative" style="min-width: 100px;">
          <div class="progress-bar progress-bar-striped" role="progressbar" style="width: {rate}%;background-color: {color};" aria-valuenow="{rate}" aria-valuemin="0" aria-valuemax="100"></div>
              <span class='badge position-absolute' style="left:auto;right:auto;width: 100%;font-size:1em">{u} {_('user.home.usage.from')} {t}</span>

        </div>
        """)    

    column_formatters = {
        'name': _name_formatter,
        'online_users': _online_users_formatter,
        'max_users': _max_users_formatter,
        'max_active_users': _max_active_users_formatter,
        'UserLinks':_ul_formatter
        
    }
    def search_placeholder(self):
        return f"{_('search')} {_('user.UUID')} {_('user.name')}"


    # def is_accessible(self):
    #     return g.admin.mode==AdminMode.super_admin


    def get_query(self):
            # Get the base query
        query = super().get_query()

        admin_ids=g.admin.recursive_sub_admins_ids()
        query = query.filter(AdminUser.id.in_(admin_ids))

        return query

    # Override get_count_query() to include the filter condition in the count query
    def get_count_query(self):
        # Get the base count query
        query = super().get_count_query()

        admin_ids=g.admin.recursive_sub_admins_ids()
        query = query.filter(AdminUser.id.in_(admin_ids))

        return query


    def on_model_change(self, form, model, is_created):
        
        # if model.id==1:
        #     model.parent_admin_id=0
        #     model.parent_admin=None
        # else:
        #     model.parent_admin_id=1
        #     model.parent_admin=AdminUser.query.filter(AdminUser.id==1).first()
        if model.id!=1 and model.parent_admin==None:
            model.parent_admin_id=g.admin.id
            model.parent_admin=g.admin

        if g.admin.mode!=AdminMode.super_admin and model.mode==AdminMode.super_admin:
            raise ValidationError("Sub-Admin can not have more power!!!!")
        if model.mode==AdminMode.agent and model.mode!=AdminMode.agent:
            raise ValidationError("Sub-Admin can not have more power!!!!")
        
    def on_model_delete(self, model):
        if model.id==1 or model.id==g.admin.id:
            raise ValidationError(_("Owner can not be deleted!"))
        users=model.recursive_users_query().all()
        for u in users:
            u.added_by=g.admin.id
        
        DailyUsage.query.filter(DailyUsage.admin_id.in_(model.recursive_sub_admins_ids())).update({'admin_id':g.admin.id})
        AdminUser.query.filter(AdminUser.id.in_(model.recursive_sub_admins_ids())).delete()

        db.session.commit()



    def get_query_for_parent_admin(self):
        admin_user_id = self.get_pk_value()
        sub_admins_ids = set(recursive_sub_admins_ids(AdminUser.query.get(admin_user_id)))
        return AdminUser.query.filter(AdminUser.id.in_(sub_admins_ids)).with_entities(AdminUser.id, AdminUser.name)


    def on_form_prefill(self, form, id=None):
        
        
        if g.admin.mode!=AdminMode.super_admin:
            del form.mode
            del form.can_add_admin
            
        if g.admin.id==form._obj.id:
            del form.max_users
            del form.max_active_users
            del form.comment
            del form.can_add_admin
            if getattr(form, 'mode'):
                del form.mode
        elif form._obj.mode==AdminMode.super_admin:
            del form.max_users
            del form.max_active_users
            del form.can_add_admin
        
        
        



from flask import Blueprint
from flask_restful import Api
from .tgbot import bot, register_bot, TGBotResource
from . import tgbot
from .tgmsg import SendMsgResource
from .resources import *
bp = Blueprint("api", __name__, url_prefix="/<proxy_path>/<user_secret>/api/v1")
api = Api(bp)


def init_app(app):
    tgbot.init_app(app)
    api.add_resource(TGBotResource, "/tgbot/")
    api.add_resource(SendMsgResource, "/send_msg/")
    api.add_resource(UserResource, "/user/")
    api.add_resource(AdminUserResource, "/admin/")

    # with app.app_context():
    #     register_bot()
    app.register_blueprint(bp)



import xtlsapi
from hiddifypanel.models import *
from .abstract_driver import DriverABS


class XrayApi(DriverABS):
    def get_xray_client(self):
        if hconfig(ConfigEnum.is_parent):
            return
        return xtlsapi.XrayClient('127.0.0.1', 10085)

    def get_enabled_users(self):
        if hconfig(ConfigEnum.is_parent):
            return
        xray_client = self.get_xray_client()
        users = User.query.all()
        t = "xtls"
        protocol = "vless"
        enabled = {}
        for u in users:
            uuid = u.uuid
            try:
                xray_client.add_client(t, f'{uuid}', f'{uuid}@hiddify.com', protocol=protocol, flow='xtls-rprx-vision', alter_id=0, cipher='chacha20_poly1305')
                xray_client.remove_client(t, f'{uuid}@hiddify.com')
                enabled[uuid] = 0
            except xtlsapi.exceptions.EmailAlreadyExists as e:
                enabled[uuid] = 1
            except Exception as e:
                print(f"error {e}")
                enabled[uuid] = e
        return enabled

    def get_inbound_tags(self):
        if hconfig(ConfigEnum.is_parent):
            return
        try:
            xray_client = self.get_xray_client()
            inbounds = [inb.name.split(">>>")[1] for inb in xray_client.stats_query('inbound')]
            print(f"Success in get inbound tags {inbounds}")
        except Exception as e:
            print(f"error in get inbound tags {e}")
            inbounds = []
        return list(set(inbounds))

    def add_client(self, user):
        if hconfig(ConfigEnum.is_parent):
            return
        uuid = user.uuid
        xray_client = self.get_xray_client()
        tags = self.get_inbound_tags()
        proto_map = {
            'vless': 'vless',
            'realityin': 'vless',
            'xtls': 'vless',
                    'quic': 'vless',
                    'trojan': 'trojan',
                    'vmess': 'vmess',
                    'ss': 'shadowsocks',
                    'v2ray': 'shadowsocks',
                    'kcp': 'vless',
                    'dispatcher': 'trojan',
        }

        def proto(t):
            res = '', ''
            for p, protocol in proto_map.items():
                if p in t:
                    res = p, protocol
                    break
            return res
        for t in tags:
            try:
                p, protocol = proto(t)
                if not p:
                    continue
                if (protocol == "vless" and p != "xtls" and p != "realityin") or "realityingrpc" in t:
                    xray_client.add_client(t, f'{uuid}', f'{uuid}@hiddify.com', protocol=protocol, flow='\0',)
                else:
                    xray_client.add_client(t, f'{uuid}', f'{uuid}@hiddify.com', protocol=protocol, flow='xtls-rprx-vision', alter_id=0, cipher='chacha20_poly1305')
                print(f"Success add  {uuid} {t}")
            except Exception as e:
                print(f"error in add  {uuid} {t} {e}")
                pass

    def remove_client(self, user):
        if hconfig(ConfigEnum.is_parent):
            return
        uuid = user.uuid
        xray_client = self.get_xray_client()
        tags = self.get_inbound_tags()

        for t in tags:
            try:
                xray_client.remove_client(t, f'{uuid}@hiddify.com')
                print(f"Success remove  {uuid} {t}")
            except Exception as e:
                print(f"error in remove  {uuid} {t} {e}")
                pass

    def get_usage(self, uuid, reset=False):
        if hconfig(ConfigEnum.is_parent):
            return
        xray_client = self.get_xray_client()
        d = xray_client.get_client_download_traffic(f'{uuid}@hiddify.com', reset=reset)
        u = xray_client.get_client_upload_traffic(f'{uuid}@hiddify.com', reset=reset)

        res = None
        if d is None:
            res = u
        elif u is None:
            res = d
        else:
            res = d + u
        if res:
            print(f"Xray usage {uuid} d={d} u={u} sum={res}")
        return res



from hiddifypanel.models import *
from hiddifypanel.panel.database import db
import sys
from hiddifypanel import Events
from dateutil import relativedelta
import datetime
from hiddifypanel.panel import hiddify
from hiddifypanel.panel.hiddify import get_random_domains, get_random_string
import random
import uuid
import urllib
import string
from hiddifypanel import Events
import random
import uuid
import urllib
import string


def init_db():
    Events.db_prehook.notify()
    db.create_all()
    hconfig.invalidate_all()
    get_hconfigs.invalidate_all()
    execute(f'update domain set mode="sub_link_only", sub_link_only=false where sub_link_only = true or mode=1  or mode="1"')
    execute(f'update domain set mode="direct", sub_link_only=false where mode=0  or mode="0"')
    execute(f'update proxy set transport="WS" where transport = "ws"')
    execute(f'update admin_user set mode="agent" where mode = "slave"')
    execute(f'update admin_user set mode="super_admin" where id=1')
    execute(f'DELETE from proxy where transport = "h1"')

    add_column(Domain.grpc)
    add_column(ParentDomain.alias)
    add_column(User.ed25519_private_key)
    add_column(User.ed25519_public_key)
    add_column(User.start_date)
    add_column(User.package_days)
    add_column(User.telegram_id)
    add_column(Child.unique_id)
    add_column(Domain.alias)
    add_column(Domain.sub_link_only)
    add_column(Domain.child_id)
    add_column(Proxy.child_id)
    add_column(User.added_by)
    add_column(User.max_ips)
    add_column(AdminUser.parent_admin_id)
    add_column(AdminUser.can_add_admin)
    add_column(AdminUser.max_active_users)
    add_column(AdminUser.max_users)
    add_column(BoolConfig.child_id)
    add_column(StrConfig.child_id)
    add_column(DailyUsage.admin_id)
    add_column(DailyUsage.child_id)
    add_column(User.monthly)
    add_column(User.enable)
    add_column(Domain.cdn_ip)
    add_column(Domain.servernames)

    if len(Domain.query.all()) != 0 and BoolConfig.query.count() == 0:
        execute(f'DROP TABLE bool_config')
        execute(f'ALTER TABLE bool_config_old RENAME TO bool_config')
    if len(Domain.query.all()) != 0 and StrConfig.query.count() == 0:
        execute(f'DROP TABLE str_config')
        execute(f'ALTER TABLE str_config_old RENAME TO str_config')

    execute('ALTER TABLE user RENAME COLUMN monthly_usage_limit_GB TO usage_limit_GB')
    execute(f'update admin_user set parent_admin_id=1 where parent_admin_id is NULL and 1!=id')
    execute(f'update admin_user set max_users=100,max_active_users=100 where max_users is NULL')
    execute(f'update dailyusage set child_id=0 where child_id is NULL')
    execute(f'update dailyusage set admin_id=1 where admin_id is NULL')
    execute(f'update dailyusage set admin_id=1 where admin_id = 0')
    execute(f'update user set added_by=1 where added_by = 0')
    execute(f'update user set enable=True, mode="no_reset" where enable is NULL')
    execute(f'update user set enable=False, mode="no_reset" where mode = "disable"')
    execute(f'update user set added_by=1 where added_by is NULL')
    execute(f'update user set max_ips=10000 where max_ips is NULL')
    execute(f'update str_config set child_id=0 where child_id is NULL')
    execute(f'update bool_config set child_id=0 where child_id is NULL')
    execute(f'update domain set child_id=0 where child_id is NULL')
    execute(f'update domain set sub_link_only=False where sub_link_only is NULL')
    execute(f'update proxy set child_id=0 where child_id is NULL')

    db_version = int(hconfig(ConfigEnum.db_version) or 0)
    start_version = db_version
    # print(f"Current DB version is {db_version}")
    if not Child.query.filter(Child.id == 0).first():
        print(Child.query.filter(Child.id == 0).first())
        db.session.add(Child(unique_id="self", id=0))
        db.session.commit()

    for ver in range(1, 60):
        if ver <= db_version:
            continue

        db_action = sys.modules[__name__].__dict__.get(f'_v{ver}', None)
        if not db_action:
            continue
        if start_version == 0 and ver == 10:
            continue

        print(f"Updating db from version {db_version}")
        db_action()
        Events.db_init_event.notify(db_version=db_version)
        print(f"Updated successfuly db from version {db_version} to {ver}")

        db_version = ver
        db.session.commit()
        set_hconfig(ConfigEnum.db_version, db_version, commit=False)

    db.session.commit()
    return BoolConfig.query.all()


# def _v50():
#     add_config_if_not_exist(ConfigEnum.tuic_enable, True)
#     add_config_if_not_exist(ConfigEnum.tuic_port, random.randint(5000, 20000))
#     if not Proxy.query.filter(Proxy.name == "TUIC").first():
#         db.session.add(Proxy(l3='custom', transport='custom', cdn='direct', proto='tuic', enable=True, name="TUIC"))

#     if not Proxy.query.filter(Proxy.name == "Hysteria").first():
#         db.session.add(Proxy(l3='custom', transport='custom', cdn='direct', proto='hysteria', enable=True, name="Hysteria"))

#     add_config_if_not_exist(ConfigEnum.hysteria_enable, True)
#     add_config_if_not_exist(ConfigEnum.hysteria_port, random.randint(5000, 20000))


def _v51():
    Proxy.query.filter(Proxy.l3.in_([ProxyL3.h3_quic])).delete()


def _v50():
    set_hconfig(ConfigEnum.show_usage_in_sublink, True)


def _v49():

    for u in User.query.all():
        priv, publ = hiddify.get_ed25519_private_public_pair()
        u.ed25519_private_key = priv
        u.ed25519_public_key = publ


def _v48():
    add_config_if_not_exist(ConfigEnum.ssh_server_enable, True)
    set_hconfig(ConfigEnum.ssh_server_enable, True)


def _v47():
    StrConfig.query.filter(StrConfig.key == ConfigEnum.ssh_server_enable).delete()


def _v45():

    if not Proxy.query.filter(Proxy.name == "SSH").first():
        db.session.add(Proxy(l3='ssh', transport='ssh', cdn='direct', proto='ssh', enable=True, name="SSH"))
    add_config_if_not_exist(ConfigEnum.ssh_server_redis_url, "unix:///opt/hiddify-config/other/redis/run.sock?db=1")
    add_config_if_not_exist(ConfigEnum.ssh_server_port, random.randint(5000, 20000))
    add_config_if_not_exist(ConfigEnum.ssh_server_enable, False)
# def _v43():
#     if not (Domain.query.filter(Domain.domain==hconfig(ConfigEnum.domain_fronting_domain)).first()):
#         db.session.add(Domain(domain=hconfig(ConfigEnum.domain_fronting_domain),servernames=hconfig(ConfigEnum.domain_fronting_domain),mode=DomainType.cdn))

# v7.0.0


def _v42():

    for k in [ConfigEnum.telegram_fakedomain, ConfigEnum.ssfaketls_fakedomain, ConfigEnum.shadowtls_fakedomain]:
        if not hconfig(k):
            rnd_domains = get_random_domains(1)
            add_config_if_not_exist(k, rnd_domains[0])


def _v41():
    add_config_if_not_exist(ConfigEnum.core_type, "xray")
    if not (Domain.query.filter(Domain.domain == hconfig(ConfigEnum.reality_fallback_domain)).first()):
        db.session.add(Domain(domain=hconfig(ConfigEnum.reality_fallback_domain), servernames=hconfig(ConfigEnum.reality_server_names), mode=DomainType.reality))


def _v38():
    add_config_if_not_exist(ConfigEnum.dns_server, "1.1.1.1")
    add_config_if_not_exist(ConfigEnum.warp_mode, "all" if hconfig(ConfigEnum.warp_enable) else "none")
    add_config_if_not_exist(ConfigEnum.warp_plus_code, '')


# def _v34():
#     add_config_if_not_exist(ConfigEnum.show_usage_in_sublink, True)


def _v33():
    Proxy.query.filter(Proxy.l3 == ProxyL3.reality).delete()
    _v31()


def _v31():
    add_config_if_not_exist(ConfigEnum.reality_short_ids, uuid.uuid4().hex[0:random.randint(1, 8)*2])
    key_pair = hiddify.generate_x25519_keys()
    add_config_if_not_exist(ConfigEnum.reality_private_key, key_pair['private_key'])
    add_config_if_not_exist(ConfigEnum.reality_public_key, key_pair['public_key'])
    db.session.bulk_save_objects(get_proxy_rows_v1())
    if not (AdminUser.query.filter(AdminUser.id == 1).first()):
        db.session.add(AdminUser(id=1, uuid=hconfig(ConfigEnum.admin_secret), name="Owner", mode=AdminMode.super_admin, comment=""))
    for i in range(1, 10):
        for d in get_random_domains(50):
            if hiddify.is_domain_reality_friendly(d):
                add_config_if_not_exist(ConfigEnum.reality_fallback_domain, d)
                add_config_if_not_exist(ConfigEnum.reality_server_names, d)
                return
    add_config_if_not_exist(ConfigEnum.reality_fallback_domain, "yahoo.com")
    add_config_if_not_exist(ConfigEnum.reality_server_names, "yahoo.com")

    # add_config_if_not_exist(ConfigEnum.cloudflare, "")


def _v27():
    # add_config_if_not_exist(ConfigEnum.cloudflare, "")
    set_hconfig(ConfigEnum.netdata, False)


def _v26():
    add_config_if_not_exist(ConfigEnum.cloudflare, "")
    add_config_if_not_exist(ConfigEnum.country, "ir")
    add_config_if_not_exist(ConfigEnum.parent_panel, "")
    add_config_if_not_exist(ConfigEnum.is_parent, False)
    add_config_if_not_exist(ConfigEnum.license, "")


def _v21():
    db.session.bulk_save_objects(get_proxy_rows_v1())


def _v20():
    if hconfig(ConfigEnum.domain_fronting_domain):
        fake_domains = [hconfig(ConfigEnum.domain_fronting_domain)]

        direct_domain = Domain.query.filter(Domain.mode in [DomainType.direct, DomainType.relay]).first()
        if direct_domain:
            direct_host = direct_domain.domain
        else:
            direct_host = hiddify.get_ip(4)

        for fd in fake_domains:
            if not Domain.query.filter(Domain.domain == fd).first():
                db.session.add(Domain(domain=fd, mode='fake', alias='moved from domain fronting', cdn_ip=direct_host))


def _v19():
    set_hconfig(ConfigEnum.path_trojan, get_random_string(7, 15))
    set_hconfig(ConfigEnum.path_vless, get_random_string(7, 15))
    set_hconfig(ConfigEnum.path_vmess, get_random_string(7, 15))
    set_hconfig(ConfigEnum.path_ss, get_random_string(7, 15))
    set_hconfig(ConfigEnum.path_grpc, get_random_string(7, 15))
    set_hconfig(ConfigEnum.path_tcp, get_random_string(7, 15))
    set_hconfig(ConfigEnum.path_ws, get_random_string(7, 15))
    add_config_if_not_exist(ConfigEnum.tuic_enable, False)
    add_config_if_not_exist(ConfigEnum.shadowtls_enable, False)
    add_config_if_not_exist(ConfigEnum.shadowtls_fakedomain, "en.wikipedia.org")
    add_config_if_not_exist(ConfigEnum.utls, "chrome")
    add_config_if_not_exist(ConfigEnum.telegram_bot_token, "")
    add_config_if_not_exist(ConfigEnum.package_mode, "release")


def _v17():
    for u in User.query.all():
        if u.expiry_time:
            if not u.package_days:
                if not u.last_reset_time:
                    u.package_days = (u.expiry_time-datetime.date.today()).days
                    u.start_date = datetime.date.today()
                else:
                    u.package_days = (u.expiry_time-u.last_reset_time).days
                    u.start_date = u.last_reset_time
            u.expiry_time = None


def _v12():
    add_column(User.last_online)
    execute(f'drop TABLE child')
    db.create_all()
    db.session.add(Child(id=0, unique_id="default"))


def _v1():
    next10year = datetime.date.today() + relativedelta.relativedelta(years=6)
    external_ip = hiddify.get_ip(4)
    rnd_domains = get_random_domains(5)
    data = [
        StrConfig(key=ConfigEnum.db_version, value=1),
        User(name="default", usage_limit_GB=3000, package_days=3650, mode=UserMode.weekly),
        Domain(domain=external_ip+".sslip.io", mode=DomainType.direct),
        StrConfig(key=ConfigEnum.admin_secret, value=str(uuid.uuid4())),
        StrConfig(key=ConfigEnum.http_ports, value="80"),
        StrConfig(key=ConfigEnum.tls_ports, value="443"),
        BoolConfig(key=ConfigEnum.first_setup, value=True),
        StrConfig(key=ConfigEnum.decoy_domain, value=rnd_domains[0]),
        StrConfig(key=ConfigEnum.proxy_path, value=get_random_string()),
        BoolConfig(key=ConfigEnum.firewall, value=False),
        BoolConfig(key=ConfigEnum.netdata, value=True),
        StrConfig(key=ConfigEnum.lang, value='en'),
        BoolConfig(key=ConfigEnum.block_iran_sites, value=True),
        BoolConfig(key=ConfigEnum.allow_invalid_sni, value=True),
        BoolConfig(key=ConfigEnum.kcp_enable, value=False),
        StrConfig(key=ConfigEnum.kcp_ports, value="88"),
        BoolConfig(key=ConfigEnum.auto_update, value=True),
        BoolConfig(key=ConfigEnum.speed_test, value=True),
        BoolConfig(key=ConfigEnum.only_ipv4, value=False),
        BoolConfig(key=ConfigEnum.vmess_enable, value=True),
        BoolConfig(key=ConfigEnum.http_proxy_enable, value=True),
        StrConfig(key=ConfigEnum.shared_secret, value=str(uuid.uuid4())),
        BoolConfig(key=ConfigEnum.telegram_enable, value=True),
        # StrConfig(key=ConfigEnum.telegram_secret,value=uuid.uuid4().hex),
        StrConfig(key=ConfigEnum.telegram_adtag, value=""),
        StrConfig(key=ConfigEnum.telegram_fakedomain, value=rnd_domains[1]),
        BoolConfig(key=ConfigEnum.ssfaketls_enable, value=False),
        # StrConfig(key=ConfigEnum.ssfaketls_secret,value=str(uuid.uuid4())),
        StrConfig(key=ConfigEnum.ssfaketls_fakedomain, value=rnd_domains[2]),
        BoolConfig(key=ConfigEnum.shadowtls_enable, value=False),
        # StrConfig(key=ConfigEnum.shadowtls_secret,value=str(uuid.uuid4())),
        StrConfig(key=ConfigEnum.shadowtls_fakedomain, value=rnd_domains[3]),

        BoolConfig(key=ConfigEnum.ssr_enable, value=False),
        # StrConfig(key=ConfigEnum.ssr_secret,value=str(uuid.uuid4())),
        StrConfig(key=ConfigEnum.ssr_fakedomain, value=rnd_domains[4]),

        BoolConfig(key=ConfigEnum.tuic_enable, value=False),
        StrConfig(key=ConfigEnum.tuic_port, value=3048),

        BoolConfig(key=ConfigEnum.domain_fronting_tls_enable, value=False),
        BoolConfig(key=ConfigEnum.domain_fronting_http_enable, value=False),
        StrConfig(key=ConfigEnum.domain_fronting_domain, value=""),

        # BoolConfig(key=ConfigEnum.torrent_block,value=False),

        *get_proxy_rows_v1()
    ]
    # fake_domains=['speedtest.net']
    # for fd in fake_domains:
    #     if not Domain.query.filter(Domain.domain==fd).first():
    #         db.session.add(Domain(domain=fd,mode='fake',alias='fake domain',cdn_ip=external_ip))
    db.session.bulk_save_objects(data)


def _v7():
    try:
        Proxy.query.filter(Proxy.name == 'tls XTLS direct trojan').delete()
        Proxy.query.filter(Proxy.name == 'tls XTLSVision direct trojan').delete()
    except:
        pass
    add_config_if_not_exist(ConfigEnum.telegram_lib, "python")
    add_config_if_not_exist(ConfigEnum.admin_lang, hconfig(ConfigEnum.lang))
    add_config_if_not_exist(ConfigEnum.branding_title, "")
    add_config_if_not_exist(ConfigEnum.branding_site, "")
    add_config_if_not_exist(ConfigEnum.branding_freetext, "")
    add_config_if_not_exist(ConfigEnum.v2ray_enable, False)
    add_config_if_not_exist(ConfigEnum.is_parent, False)
    add_config_if_not_exist(ConfigEnum.parent_panel, '')
    add_config_if_not_exist(ConfigEnum.unique_id, str(uuid.uuid4()))


def _v9():
    # add_column(User.mode)
    # add_column(User.comment)
    try:
        for u in User.query.all():
            u.mode = UserMode.monthly if u.monthly else UserMode.no_reset
    except:
        pass


def _v10():
    all_configs = get_hconfigs()
    execute("ALTER TABLE `str_config` RENAME TO `str_config_old`")
    execute("ALTER TABLE `bool_config` RENAME TO `bool_config_old`")
    db.create_all()
    rows = []
    for c, v in all_configs.items():
        if c.type() == bool:
            rows.append(BoolConfig(key=c, value=v, child_id=0))
        else:
            rows.append(StrConfig(key=c, value=v, child_id=0))

    db.session.bulk_save_objects(rows)


def get_proxy_rows_v1():
    return make_proxy_rows([
        # 'WS Fake vless',
        # 'WS Fake trojan',
        # 'WS Fake vmess',
        # 'grpc Fake vless',
        # 'grpc Fake trojan',
        # 'grpc Fake vmess',
        # "XTLS direct vless",
        # "XTLS direct trojan",
        "XTLS direct vless",
        "WS direct vless",
        "WS direct trojan",
        "WS direct vmess",
        "WS CDN vless",
        "WS CDN trojan",
        "WS CDN vmess",
        "grpc CDN vless",
        "grpc CDN trojan",
        "grpc CDN vmess",
        "tcp direct vless",
        "tcp direct trojan",
        "tcp direct vmess",
        "grpc direct vless",
        "grpc direct trojan",
        "grpc direct vmess",
        # "h1 direct vless",
        # "h1 direct vmess",
        "faketls direct ss",
        "WS direct v2ray",
        "shadowtls direct ss",
        "restls1_2 direct ss",
        "restls1_3 direct ss",
        "tcp direct ssr",
        "WS CDN v2ray"
    ]
    )


def make_proxy_rows(cfgs):
    # "h3_quic",
    for l3 in ["tls_h2", "tls", "http", "kcp", "reality"]:
        for c in cfgs:
            transport, cdn, proto = c.split(" ")
            if l3 in ["kcp", 'reality'] and cdn != "direct":
                continue
            if l3 == "reality" and ((transport not in ['tcp', 'grpc', 'XTLS']) or proto != 'vless'):
                continue
            if proto == "trojan" and l3 not in ["tls", 'xtls', 'tls_h2', 'h3_quic']:
                continue
            if transport in ["grpc", "XTLS", "faketls"] and l3 == "http":
                continue
            # if l3 == "tls_h2" and transport =="grpc":
            #     continue
            enable = l3 != "http" or proto == "vmess"
            enable = enable and transport != 'tcp'
            name = f'{l3} {c}'
            is_exist = Proxy.query.filter(Proxy.name == name).first() or Proxy.query.filter(
                Proxy.l3 == l3, Proxy.transport == transport, Proxy.cdn == cdn, Proxy.proto == proto).first()
            if not is_exist:
                yield Proxy(l3=l3, transport=transport, cdn=cdn, proto=proto, enable=enable, name=name)


def add_config_if_not_exist(key: ConfigEnum, val):
    table = BoolConfig if key.type() == bool else StrConfig
    if table.query.filter(table.key == key).count() == 0:
        db.session.add(table(key=key, value=val, child_id=0))


def add_column(column):
    try:
        column_type = column.type.compile(db.engine.dialect)
        db.engine.execute(f'ALTER TABLE {column.table.name} ADD COLUMN {column.name} {column_type}')
    except:
        pass


def execute(query):
    try:
        db.engine.execute(query)
    except:
        pass



from hiddifypanel.panel import hiddify
from sqlalchemy.orm import Load, joinedload


from hiddifypanel.drivers import user_driver
from sqlalchemy import func
from flask_babelex import gettext as __
from flask_babelex import lazy_gettext as _
import urllib
from hiddifypanel.models import *
from hiddifypanel.panel.database import db
import datetime
from flask import jsonify, g, url_for, Markup
from flask import flash as flask_flash
to_gig_d = 1024**3


def update_local_usage():
    res = user_driver.get_users_usage(reset=True)
    return add_users_usage(res, child_id=0)

    # return {"status": 'success', "comments":res}


def add_users_usage_uuid(uuids_bytes, child_id):
    users = User.query.filter(User.uuid.in_(keys(uuids_bytes)))
    dbusers_bytes = {u: uuids_bytes.get(u.uuid, 0) for u in users}
    add_users_usage(dbusers_bytes, child_id)


def add_users_usage(dbusers_bytes, child_id):
    print(dbusers_bytes)
    if not hconfig(ConfigEnum.is_parent) and hconfig(ConfigEnum.parent_panel):
        from hiddifypanel.panel import hiddify_api
        hiddify_api.add_user_usage_to_parent(dbusers_bytes)

    res = {}
    have_change = False
    before_enabled_users = user_driver.get_enabled_users()
    daily_usage = {}
    today = datetime.date.today()
    for adm in AdminUser.query.all():
        daily_usage[adm.id] = DailyUsage.query.filter(DailyUsage.date == today, DailyUsage.admin_id == adm.id, DailyUsage.child_id == child_id).first()
        if not daily_usage[adm.id]:
            daily_usage[adm.id] = DailyUsage(admin_id=adm.id, child_id=child_id)
            db.session.add(daily_usage[adm.id])
        daily_usage[adm.id].online = User.query.filter(User.added_by == adm.id).filter(func.DATE(User.last_online) == today).count()
    # db.session.commit()
    userDetails = {p.user_id: p for p in UserDetail.query.filter(UserDetail.child_id == child_id).all()}
    for user, uinfo in dbusers_bytes.items():
        usage_bytes = uinfo['usage']
        ips = uinfo['ips']
        # user_active_before=is_user_active(user)
        detail = userDetails.get(user.id)
        if not detail:
            detail = UserDetail(user_id=user.id, child_id=child_id)
            detail.current_usage_GB = detail.current_usage_GB or 0
            db.session.add(detail)
        detail.connected_ips = ips
        detail.current_usage_GB = detail.current_usage_GB or 0
        if not user.last_reset_time or user_should_reset(user):
            user.last_reset_time = datetime.date.today()
            user.current_usage_GB = 0
            detail.current_usage_GB = 0

        if not before_enabled_users[user.uuid] and user.is_active:
            print("Enabling disabled client {user.uuid} ")
            user_driver.add_client(user)
            send_bot_message(user)
            have_change = True
        if type(usage_bytes) != int or usage_bytes == 0:
            res[user.uuid] = "No usage"
        else:
            daily_usage.get(user.added_by, daily_usage[1]).usage += usage_bytes
            in_gig = (usage_bytes)/to_gig_d
            res[user.uuid] = in_gig
            user.current_usage_GB += in_gig
            detail.current_usage_GB += in_gig
            user.last_online = datetime.datetime.now()
            detail.last_online = datetime.datetime.now()

            if user.start_date == None:
                user.start_date = datetime.date.today()

        if before_enabled_users[user.uuid] and not is_user_active(user):
            print("Removing enabled client {user.uuid} ")
            user_driver.remove_client(user)
            have_change = True
            res[user.uuid] = f"{res[user.uuid]} !OUT of USAGE! Client Removed"

    db.session.commit()
    if have_change and hconfig(ConfigEnum.core_type == 'singbox'):
        hiddify.quick_apply_users()

    return {"status": 'success', "comments": res}


def send_bot_message(user):
    if not (hconfig(ConfigEnum.telegram_bot_token) and not hconfig(ConfigEnum.parent_panel)):
        return
    if not user.telegram_id:
        return
    from hiddifypanel.panel.commercial.telegrambot import bot, Usage
    try:
        msg = Usage.get_usage_msg(user.uuid)
        msg = _("User activated!") if is_user_active(user) else _("Package ended!")+"\n"+msg
        bot.send_message(user.telegram_id, msg, reply_markup=Usage.user_keyboard(uuid))
    except:
        pass



from flask_admin.contrib import sqla
from hiddifypanel.panel.database import db
from wtforms.validators import Regexp
from hiddifypanel.models import *
from wtforms.validators import Regexp, ValidationError
from hiddifypanel.panel.admin.adminlte import AdminLTEModelView
from flask_babelex import gettext as __
from flask_babelex import lazy_gettext as _
from hiddifypanel.panel import hiddify
from flask import Markup, g
from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from flask_admin import Admin
from flask_admin.contrib.sqla import ModelView
from wtforms import SelectMultipleField


from wtforms.widgets import ListWidget, CheckboxInput
from sqlalchemy.orm import backref
# Define a custom field type for the related domains
from flask_admin.form.fields import Select2TagsField, Select2Field


class ProxyDetailsAdmin(AdminLTEModelView):
    column_hide_backrefs = True
    form_excluded_columns = ['child']
    column_exclude_list = ['child']
    # list_template = 'model/domain_list.html'
    # edit_modal = True
    # form_overrides = {'work_with': Select2Field}

    def after_model_change(self, form, model, is_created):
        # if hconfig(ConfigEnum.parent_panel):
        #     hiddify_api.sync_child_to_parent()
        hiddify.get_available_proxies.invalidate_all()
        pass

    def after_model_delete(self, model):
        # if hconfig(ConfigEnum.parent_panel):
        #     hiddify_api.sync_child_to_parent()
        pass

    def is_accessible(self):
        return g.admin.mode in [AdminMode.admin, AdminMode.super_admin]



from sqlalchemy_serializer import SerializerMixin
from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from flask_admin import Admin
from flask_admin.contrib.sqla import ModelView
from wtforms import SelectMultipleField

from sqlalchemy.orm import backref

from flask import Markup
from dateutil import relativedelta
from hiddifypanel.panel.database import db
import enum
import datetime
import uuid as uuid_mod
from enum import auto
from strenum import StrEnum


class ConfigCategory(StrEnum):
    admin = auto()
    branding = auto()
    general = auto()
    proxies = auto()
    domain_fronting = auto()
    telegram = auto()
    http = auto()
    tls = auto()
    ssh = auto()
    ssfaketls = auto()
    shadowtls = auto()
    restls = auto()
    tuic = auto()
    hysteria = auto()
    ssr = auto()
    kcp = auto()
    hidden = auto()
    advanced = auto()
    too_advanced = auto()
    warp = auto()
    reality = auto()


class ConfigEnum(StrEnum):
    ssh_server_redis_url = auto()
    ssh_server_port = auto()
    ssh_server_enable = auto()
    first_setup = auto()
    core_type = auto()
    warp_enable = auto()
    warp_mode = auto()
    warp_plus_code = auto()
    dns_server = auto()
    reality_fallback_domain = auto()
    reality_server_names = auto()
    reality_short_ids = auto()
    reality_private_key = auto()
    reality_public_key = auto()

    restls1_2_domain = auto()
    restls1_3_domain = auto()
    show_usage_in_sublink = auto()
    cloudflare = auto()
    license = auto()
    country = auto()
    package_mode = auto()
    utls = auto()
    telegram_bot_token = auto()
    is_parent = auto()
    parent_panel = auto()
    unique_id = auto()
    last_hash = auto()
    cdn_forced_host = auto()  # removed
    lang = auto()
    admin_lang = auto()
    admin_secret = auto()
    tls_ports = auto()
    http_ports = auto()
    kcp_ports = auto()
    kcp_enable = auto()
    decoy_domain = auto()
    proxy_path = auto()
    firewall = auto()
    netdata = auto()
    http_proxy_enable = auto()
    block_iran_sites = auto()
    allow_invalid_sni = auto()
    auto_update = auto()
    speed_test = auto()
    only_ipv4 = auto()

    shared_secret = auto()

    telegram_enable = auto()
    # telegram_secret=auto()
    telegram_adtag = auto()
    telegram_lib = auto()
    telegram_fakedomain = auto()

    v2ray_enable = auto()
    torrent_block = auto()

    ssfaketls_enable = auto()
    # ssfaketls_secret="ssfaketls_secret"
    ssfaketls_fakedomain = auto()

    shadowtls_enable = auto()
    # shadowtls_secret=auto()
    shadowtls_fakedomain = auto()

    tuic_enable = auto()
    tuic_port = auto()

    hysteria_enable = auto()
    hysteria_port = auto()

    ssr_enable = auto()
    # ssr_secret="ssr_secret"
    ssr_fakedomain = auto()

    vmess_enable = auto()
    domain_fronting_domain = auto()
    domain_fronting_http_enable = auto()
    domain_fronting_tls_enable = auto()

    db_version = auto()

    branding_title = auto()
    branding_site = auto()
    branding_freetext = auto()
    not_found = auto()
    path_vmess = auto()
    path_vless = auto()
    path_trojan = auto()
    path_v2ray = auto()  # deprecated
    path_ss = auto()

    path_ws = auto()
    path_tcp = auto()
    path_grpc = auto()

    # cdn_forced_host=auto()
    @ classmethod
    def _missing_(cls, value):
        return cls.not_found  # "key not found"

    def info(self):
        map = {
            self.ssh_server_redis_url: {'category': ConfigCategory.hidden},
            self.ssh_server_port: {'category': ConfigCategory.ssh, 'apply_mode': 'apply'},
            self.ssh_server_enable: {'category': ConfigCategory.ssh, 'type': bool, 'apply_mode': 'apply'},
            self.core_type: {'category': ConfigCategory.advanced, 'apply_mode': 'apply'},
            self.dns_server: {'category': ConfigCategory.general, 'apply_mode': 'apply'},
            self.warp_enable: {'category': ConfigCategory.hidden, 'type': bool, 'apply_mode': 'restart'},
            self.warp_mode: {'category': ConfigCategory.warp, 'apply_mode': 'apply'},
            self.warp_plus_code: {'category': ConfigCategory.warp, 'apply_mode': 'apply'},
            self.restls1_2_domain: {'category': ConfigCategory.hidden},
            self.restls1_3_domain: {'category': ConfigCategory.hidden},
            self.first_setup: {'category': ConfigCategory.hidden, 'type': bool},
            self.show_usage_in_sublink: {'category': ConfigCategory.general, 'type': bool},
            self.reality_fallback_domain: {'category': ConfigCategory.hidden, 'apply_mode': 'apply'},
            self.reality_server_names: {'category': ConfigCategory.hidden, 'apply_mode': 'apply'},
            self.reality_short_ids: {'category': ConfigCategory.reality, 'apply_mode': 'apply'},
            self.reality_private_key: {'category': ConfigCategory.reality, 'apply_mode': 'apply'},
            self.reality_public_key: {'category': ConfigCategory.reality, 'apply_mode': 'apply'},
            self.lang: {'category': ConfigCategory.branding, 'show_in_parent': True, 'commercial': True},
            self.admin_lang: {'category': ConfigCategory.admin, 'show_in_parent': True, 'commercial': True},
            self.cloudflare: {'category': ConfigCategory.too_advanced, 'commercial': True},
            self.license: {'category': ConfigCategory.hidden, 'commercial': True},
            self.proxy_path: {'category': ConfigCategory.too_advanced, 'apply_mode': 'apply', 'show_in_parent': True},
            self.path_vmess: {'category': ConfigCategory.too_advanced},
            self.path_vless: {'category': ConfigCategory.too_advanced},
            self.path_trojan: {'category': ConfigCategory.too_advanced},
            self.path_ss: {'category': ConfigCategory.hidden},
            self.path_tcp: {'category': ConfigCategory.too_advanced},
            self.path_ws: {'category': ConfigCategory.too_advanced},
            self.path_grpc: {'category': ConfigCategory.too_advanced},
            self.path_v2ray: {'category': ConfigCategory.hidden},
            self.last_hash: {'category': ConfigCategory.hidden},

            self.utls: {'category': ConfigCategory.advanced, 'commercial': True},
            self.package_mode: {'category': ConfigCategory.advanced, 'show_in_parent': True, 'commercial': True},
            self.telegram_bot_token: {'category': ConfigCategory.telegram, 'show_in_parent': True, 'commercial': True},
            self.is_parent: {'category': ConfigCategory.hidden, 'type': bool},
            self.parent_panel: {'category': ConfigCategory.hidden, 'commercial': True},
            self.unique_id: {'category': ConfigCategory.hidden, },
            self.cdn_forced_host: {'category': ConfigCategory.hidden, },
            self.branding_title: {'category': ConfigCategory.branding, 'show_in_parent': True, 'commercial': True},
            self.branding_site: {'category': ConfigCategory.branding, 'show_in_parent': True, 'commercial': True},
            self.branding_freetext: {'category': ConfigCategory.branding, 'show_in_parent': True, 'commercial': True},
            self.not_found: {'category': ConfigCategory.hidden},
            self.admin_secret: {'category': ConfigCategory.hidden, 'show_in_parent': True, 'commercial': True},

            self.tls_ports: {'category': ConfigCategory.tls, 'apply_mode': 'apply'},  # tls
            self.http_ports: {'category': ConfigCategory.http, 'apply_mode': 'apply'},  # http
            self.kcp_ports: {'category': ConfigCategory.hidden, 'apply_mode': 'apply'},
            self.kcp_enable: {'category': ConfigCategory.hidden, 'type': bool, 'apply_mode': 'apply'},
            self.decoy_domain: {'category': ConfigCategory.general, 'apply_mode': 'apply'},
            self.country: {'category': ConfigCategory.general, 'apply_mode': 'restart'},
            self.firewall: {'category': ConfigCategory.general, 'type': bool, 'apply_mode': 'apply'},
            self.netdata: {'category': ConfigCategory.hidden, 'type': bool, 'apply_mode': 'apply'},
            self.http_proxy_enable: {'category': ConfigCategory.http, 'type': bool},
            self.block_iran_sites: {'category': ConfigCategory.proxies, 'type': bool, 'apply_mode': 'apply'},
            self.allow_invalid_sni: {'category': ConfigCategory.tls, 'type': bool, 'apply_mode': 'apply'},
            self.auto_update: {'category': ConfigCategory.general, 'type': bool, 'apply_mode': 'apply', 'show_in_parent': True},
            self.speed_test: {'category': ConfigCategory.general, 'type': bool, 'apply_mode': 'apply'},
            self.only_ipv4: {'category': ConfigCategory.general, 'type': bool, 'apply_mode': 'apply'},
            self.torrent_block: {'category': ConfigCategory.general, 'type': bool, 'apply_mode': 'apply'},

            self.shared_secret: {'category': ConfigCategory.proxies, 'apply_mode': 'apply'},

            self.telegram_enable: {'category': ConfigCategory.telegram, 'type': bool, 'apply_mode': 'apply'},
            # telegram_secret:{'category':'general'},
            self.telegram_adtag: {'category': ConfigCategory.telegram, 'apply_mode': 'apply'},
            self.telegram_fakedomain: {'category': ConfigCategory.telegram, 'apply_mode': 'apply'},
            self.telegram_lib: {'category': ConfigCategory.telegram, 'apply_mode': 'reinstall'},

            self.v2ray_enable: {'category': ConfigCategory.proxies, 'type': bool, 'apply_mode': 'apply'},

            self.ssfaketls_enable: {'category': ConfigCategory.ssfaketls, 'type': bool, 'apply_mode': 'apply'},
            # ssfaketls_secret:{'category':'ssfaketls'},
            self.ssfaketls_fakedomain: {'category': ConfigCategory.ssfaketls, 'apply_mode': 'apply'},

            self.shadowtls_enable: {'category': ConfigCategory.shadowtls, 'type': bool, 'apply_mode': 'apply'},
            # shadowtls_secret:{'category':'shadowtls'},
            self.shadowtls_fakedomain: {'category': ConfigCategory.shadowtls, 'apply_mode': 'apply'},

            self.tuic_enable: {'category': ConfigCategory.hidden, 'type': bool, 'apply_mode': 'apply'},
            self.tuic_port: {'category': ConfigCategory.hidden, 'apply_mode': 'apply'},

            self.hysteria_enable: {'category': ConfigCategory.hidden, 'type': bool, 'apply_mode': 'apply'},
            self.hysteria_port: {'category': ConfigCategory.hidden, 'apply_mode': 'apply'},

            self.ssr_enable: {'category': ConfigCategory.hidden, 'type': bool, 'apply_mode': 'apply'},
            # ssr_secret:{'category':'ssr'},
            self.ssr_fakedomain: {'category': ConfigCategory.hidden, 'apply_mode': 'apply'},

            self.vmess_enable: {'category': ConfigCategory.proxies, 'type': bool},
            self.domain_fronting_domain: {'category': ConfigCategory.hidden},
            self.domain_fronting_http_enable: {'category': ConfigCategory.hidden, 'type': bool},
            self.domain_fronting_tls_enable: {'category': ConfigCategory.hidden, 'type': bool},
            self.db_version: {'category': ConfigCategory.hidden},
        }
        return map[self]

    def commercial(self):
        # print(self,self.info().get('commercial',False))
        return self.info().get('commercial', False)

    def category(self):
        return self.info()['category']

    def type(self):
        return self.info().get('type', str)

    def apply_mode(self):
        return self.info().get('apply_mode', '')

    def show_in_parent(self):
        return self.info().get('show_in_parent', False)



from redis_cache import RedisCache
from functools import wraps
import redis
from pickle import dumps, loads
redis_client = redis.from_url('unix:///opt/hiddify-config/other/redis/run.sock?db=0')


def exception_handler(e, original_fn, args, kwargs):
    return original_fn(*args, **kwargs)
    pass


# cache = RedisCache(redis_client=redis_client, exception_handler=exception_handler)
# cache = RedisCache(redis_client=redis_client, prefix="h", serializer=dumps, deserializer=loads, exception_handler=exception_handler)
cache = RedisCache(redis_client=redis_client, prefix="h", serializer=dumps, deserializer=loads)


class CacheDecorator:
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, fn):
        @wraps(fn)
        def inner(*args, **kwargs):

            parsed_result = fn(*args, **kwargs)

            return parsed_result

        inner.invalidate = self.invalidate
        inner.invalidate_all = self.invalidate_all
        inner.instance = self
        return inner

    def invalidate(self, *args, **kwargs):
        pass

    def invalidate_all(self, *args, **kwargs):
        pass


class DisableCache:
    cache = CacheDecorator


# cache = DisableCache()
try:
    @cache.cache()
    def test():
        return 1
    test()
except Exception as e:
    import sys
    print('Caching Error! Disabling cache', e, file=sys.stderr)
    cache = DisableCache()



from flask import abort, jsonify, request
from flask_restful import Resource
# from flask_simplelogin import login_required
import datetime
from hiddifypanel.models import *
from urllib.parse import urlparse
from hiddifypanel.panel import hiddify
from hiddifypanel.drivers import user_driver
# class AllResource(Resource):
#     def get(self):
#         return jsonify(
#             hiddify.dump_db_to_dict()
#         )


class UserResource(Resource):
    def get(self, uuid=None):
        if uuid:
            product = User.query.filter(User.uuid == uuid).first() or abort(204)
            return jsonify(product.to_dict())

        products = User.query.all() or abort(204)
        return jsonify(
            [product.to_dict() for product in products]
        )

    def post(self):
        data = request.json
        hiddify.add_or_update_user(**data)
        user_driver.add_client(data['uuid'])
        hiddify.quick_apply_users()

        return jsonify({'status': 200, 'msg': 'ok'})


class AdminUserResource(Resource):
    def get(self, uuid=None):
        if uuid:
            product = AdminUser.query.filter(AdminUser.uuid == uuid).first() or abort(204)
            return jsonify(product.to_dict())

        products = AdminUser.query.all() or abort(204)
        return jsonify(
            [product.to_dict() for product in products]
        )

    def post(self):
        data = request.json
        hiddify.add_or_update_admin(**data)

        return jsonify({'status': 200, 'msg': 'ok'})


# class DomainResource(Resource):
#     def get(self,domain=None):
#         if domain:
#             product = Domain.query.filter(Domain.domain==domain).first() or abort(204)
#             return jsonify(hiddify.domain_dict(product))
#         products = Domain.query.all() or abort(204)
#         return jsonify(
#             [hiddify.domain_dict(product) for product in products]
#         )
#     def post(self):
#         hiddify.add_or_update_domain(**request.json)
#         return jsonify({'status':200,'msg':'ok'})

# class ParentDomainResource(Resource):
#     def get(self,parent_domain=None):
#         if domain:
#             product = ParentDomain.query.filter(ParentDomain.domain==domain).first() or abort(204)
#             return jsonify(hiddify.parent_domain_dict(product))
#         products = ParentDomain.query.all() or abort(204)
#         return jsonify(
#             [hiddify.parent_domain_dict(product) for product in products]
#         )
#     def post(self):
#         hiddify.add_or_update_parent_domain(**request.json)
#         return jsonify({'status':200,'msg':'ok'})

# class ConfigResource(Resource):
#     def get(self,key=None,child_id=0):
#         if key:
#             return jsonify(hconfig(key,child_id))
#         return jsonify(get_hconfigs(child_id))

#     def post(self):
#         hiddify.add_or_update_config(**request.json)
#         return jsonify({'status':200,'msg':'ok'})

class HelloResource(Resource):
    def get(self):
        return jsonify({"status": 200, "msg": "ok"})


# class UpdateUsageResource(Resource):
#     def get(self):
#         return jsonify({"status": 200, "msg": "ok"})



from flask_admin.base import AdminIndexView, expose
from hiddifypanel.panel.hiddify import admin
from flask import render_template, url_for, Markup, request, jsonify, g, redirect
from flask_babelex import lazy_gettext as _
from hiddifypanel.panel import hiddify
import datetime
from flask_classful import FlaskView, route
from hiddifypanel.models import *
from hiddifypanel.panel.database import db
from hiddifypanel.panel.hiddify import flash
import hiddifypanel


class Dashboard(FlaskView):
    def get_data(self):
        admin_id = request.args.get("admin_id") or g.admin.id
        if admin_id not in g.admin.recursive_sub_admins_ids():
            abort(403, _("Access Denied!"))

        return jsonify(dict(
            stats={'system': hiddify.system_stats(), 'top5': hiddify.top_processes()},
            usage_history=get_daily_usage_stats(admin_id)
        ))

    def index(self):
        print(hconfig(ConfigEnum.first_setup))
        if hconfig(ConfigEnum.first_setup):
            return redirect(url_for("admin.QuickSetup:index"))
        if hiddifypanel.__release_date__ + datetime.timedelta(days=20) < datetime.datetime.now():
            flash(_('This version of hiddify panel is outdated. Please update it from admin area.'), "danger")
        bot = None
        # if hconfig(ConfigEnum.license):
        childs = None
        admin_id = request.args.get("admin_id") or g.admin.id
        if admin_id not in g.admin.recursive_sub_admins_ids():
            abort(403, _("Access Denied!"))

        child_id = request.args.get("child_id") or None
        user_query = User.query
        if admin_id:
            user_query = user_query.filter(User.added_by == admin_id)
        if hconfig(ConfigEnum.is_parent):
            childs = Child.query.filter(Child.id != 0).all()
            for c in childs:
                c.is_active = False
                for d in c.domains:
                    if d.mode == DomainType.fake:
                        continue
                    remote = f"https://{d.domain}/{hconfig(ConfigEnum.proxy_path,c.id)}/{hconfig(ConfigEnum.admin_secret,c.id)}"
                    d.is_active = hiddify.check_connection_to_remote(remote)
                    if d.is_active:
                        c.is_active = True

            # return render_template('parent_dash.html',childs=childs,bot=bot)
    # try:
        def_user = None if len(User.query.all()) > 1 else User.query.filter(User.name == 'default').first()
        domains = get_panel_domains()
        sslip_domains = [d.domain for d in domains if "sslip.io" in d.domain]

        if def_user and sslip_domains:
            quick_setup = url_for("admin.QuickSetup:index")
            flash((_('It seems that you have not setup the system completely. <a class="btn btn-success" href="%(quick_setup)s">Click here</a> to complete setup.', quick_setup=quick_setup)), 'warning')
            if hconfig(ConfigEnum.is_parent):
                flash(_("Please understand that parent panel is under test and the plan and the condition of use maybe change at anytime."), "danger")
        elif len(sslip_domains):
            flash((_('It seems that you are using default domain (%(domain)s) which is not recommended.', domain=sslip_domains[0])), 'warning')
            if hconfig(ConfigEnum.is_parent):
                flash(_("Please understand that parent panel is under test and the plan and the condition of use maybe change at anytime."), "danger")
        elif def_user:
            d = domains[0]
            u = def_user.uuid
            flash((_('It seems that you have not created any users yet. Default user link: %(default_link)s', default_link=hiddify.get_user_link(u, d))), 'secondary')

    # except:
    #     flash((_('Error!!!')),'info')

        stats = {'system': hiddify.system_stats(), 'top5': hiddify.top_processes()}
        return render_template('index.html', stats=stats, usage_history=get_daily_usage_stats(admin_id, child_id), childs=childs)

    @route('remove_child', methods=['POST'])
    def remove_child(self):
        child_id = request.form['child_id']
        child = Child.query.filter(Child.id == child_id).first()
        db.session.delete(child)
        db.session.commit()
        flash(_("child has been removed!"), "success")
        return self.index()



import xtlsapi
from hiddifypanel.models import *
from .abstract_driver import DriverABS


class SingboxApi(DriverABS):
    def get_singbox_client(self):
        if hconfig(ConfigEnum.is_parent):
            return
        return xtlsapi.SingboxClient('127.0.0.1', 10086)

    def get_enabled_users(self):
        if hconfig(ConfigEnum.is_parent):
            return
        raise NotImplementedError()

    def get_inbound_tags(self):
        if hconfig(ConfigEnum.is_parent):
            return
        try:
            xray_client = self.get_singbox_client()
            inbounds = [inb.name.split(">>>")[1] for inb in xray_client.stats_query('inbound')]
            print(f"Success in get inbound tags {inbounds}")
        except Exception as e:
            print(f"error in get inbound tags {e}")
            inbounds = []
        return list(set(inbounds))

    def add_client(self, user):
        if hconfig(ConfigEnum.is_parent):
            return
        # raise NotImplementedError()

    def remove_client(self, user):
        if hconfig(ConfigEnum.is_parent):
            return
        # raise NotImplementedError()

    def get_usage(self, uuid, reset=False):
        if hconfig(ConfigEnum.is_parent):
            return
        xray_client = self.get_singbox_client()
        d = xray_client.get_client_download_traffic(f'{uuid}@hiddify.com', reset=reset)
        u = xray_client.get_client_upload_traffic(f'{uuid}@hiddify.com', reset=reset)

        res = None
        if d is None:
            res = u
        elif u is None:
            res = d
        else:
            res = d + u
        if res:
            print(f"singbox {uuid} d={d} u={u} sum={res}")
        return res



from hiddifypanel.models import BoolConfig, StrConfig, ConfigEnum, hconfig, Proxy
# from flask_babelex import lazy_gettext as __
from flask_babelex import gettext as _
import wtforms as wtf
from flask_wtf import FlaskForm
from flask_bootstrap import SwitchField

from flask_admin.base import expose
# from gettext import gettext as _
from hiddifypanel.panel import hiddify


import re
from flask import render_template, current_app, Markup, url_for, abort
from hiddifypanel.panel.hiddify import flash
from hiddifypanel.models import User, Domain, DomainType, StrConfig, ConfigEnum, get_hconfigs
from hiddifypanel.panel.database import db
from wtforms.fields import *
from hiddifypanel.panel import hiddify
from flask_classful import FlaskView
from hiddifypanel.panel.hiddify import flash


class ProxyAdmin(FlaskView):

    @hiddify.admin
    def index(self):
        return render_template('proxy.html', global_config_form=get_global_config_form(), detailed_config_form=get_all_proxy_form())

    @hiddify.admin
    def post(self):
        global_config_form = get_global_config_form()
        all_proxy_form = get_all_proxy_form()

        if global_config_form.submit_global.data and global_config_form.validate_on_submit():
            old_configs = get_hconfigs()
            for k, vs in global_config_form.data.items():
                if k in [c for c in ConfigEnum]:
                    BoolConfig.query.filter(BoolConfig.key == k).first().value = vs
                    if vs and k in [ConfigEnum.domain_fronting_http_enable, ConfigEnum.domain_fronting_tls_enable] and hconfig(ConfigEnum.domain_fronting_domain) == "":
                        flash((_('config.domain-fronting-notsetup-error')), 'danger')

            # print(cat,vs)
            hiddify.get_available_proxies.invalidate_all()
            db.session.commit()
            hiddify.check_need_reset(old_configs)
            all_proxy_form = get_all_proxy_form(True)

        elif all_proxy_form.submit_detail.data and all_proxy_form.validate_on_submit():

            for cdn, vs in all_proxy_form.data.items():  # [c for c in ConfigEnum]:
                if type(vs) is not dict:
                    continue
                for proto, v in vs.items():  # [c for c in ConfigEnum]:
                    if type(v) is not dict:
                        continue
                    for proxy_id, enable in v.items():
                        if not proxy_id.startswith("p_"):
                            continue
                        id = int(proxy_id.split('_')[-1])
                        Proxy.query.filter(Proxy.id == id).first().enable = enable

                # print(cat,vs)
            db.session.commit()
            hiddify.get_available_proxies.invalidate_all()
            hiddify.flash_config_success(restart_mode='apply', domain_changed=False)
            # if hconfig(ConfigEnum.parent_panel):
            #     hiddify_api.sync_child_to_parent()
            global_config_form = get_global_config_form(True)
        else:
            flash((_('config.validation-error')), 'danger')

        return render_template('proxy.html', global_config_form=global_config_form, detailed_config_form=all_proxy_form)

        import flask_babelex

        # form=HelloForm()
        # # return render('config.html',form=form)
        # return render_template('config.html',form=HelloForm())
        form = get_config_form()
        return render_template('config.html', form=form)


def get_global_config_form(empty=False):
    boolconfigs = BoolConfig.query.all()

    class DynamicForm(FlaskForm):
        pass

    for cf in boolconfigs:
        if cf.key.category() == 'hidden':
            continue
        if not cf.key.endswith("_enable"):
            continue
        field = SwitchField(_(f'config.{cf.key}.label'), default=cf.value, description=_(f'config.{cf.key}.description'))
        setattr(DynamicForm, cf.key,  field)
    setattr(DynamicForm, "submit_global", wtf.fields.SubmitField(_('Submit')))
    if empty:
        return DynamicForm(None)
    return DynamicForm()


def get_all_proxy_form(empty=False):
    proxies = hiddify.get_available_proxies(0)
    categories1 = sorted([c for c in {c.cdn: 1 for c in proxies}])

    class DynamicForm(FlaskForm):
        pass

    for cdn in categories1:
        class CDNForm(FlaskForm):
            pass
        cdn_proxies = [c for c in proxies if c.cdn == cdn]
        protos = sorted([c for c in {c.proto: 1 for c in cdn_proxies}])
        for proto in protos:
            class ProtoForm(FlaskForm):
                pass
            proto_proxies = [c for c in cdn_proxies if c.proto == proto]
            for proxy in proto_proxies:
                field = SwitchField(proxy.name, default=proxy.enable, description=f"l3:{proxy.l3} transport:{proxy.transport}")
                setattr(ProtoForm, f"p_{proxy.id}",  field)

            multifield = wtf.fields.FormField(ProtoForm, proto)
            setattr(CDNForm, proto,  multifield)
        field_name = cdn if cdn != "Fake" else _('config.domain_fronting.label')
        multifield = wtf.fields.FormField(CDNForm, field_name)
        setattr(DynamicForm, cdn,  multifield)
    setattr(DynamicForm, "submit_detail", wtf.fields.SubmitField(_('Submit')))
    if empty:
        return DynamicForm(None)
    return DynamicForm()



from .ProxyDetailsAdmin import ProxyDetailsAdmin
# from .CommercialSettings import CommercialSettings
from hiddifypanel.panel import hiddify
from hiddifypanel.models import *
from hiddifypanel.panel.database import db
from hiddifypanel import Events

commercial = False


def init_app(app):
    from . import restapi
    restapi.init_app(app)
    return


def is_valid():
    return True


def config_changed_event(conf, old_value):
    if conf.key == ConfigEnum.is_parent:
        if conf.value and not is_valid():
            set_hconfig(ConfigEnum.is_parent, False)
        if not old_value and conf.value:
            Domain.query.delete()
            new_domain = hiddify.get_ip(4)+".sslip.io"
            if not ParentDomain.query.filter(ParentDomain.domain == new_domain).first():
                db.session.add(ParentDomain(domain=hiddify.get_ip(4)+".sslip.io"))
                db.session.commit()


Events.config_changed.subscribe(config_changed_event)


# def db_init_event(db_version):
#     # if not is_valid():
#     #     db.add(BoolConfig(ConfigEnum.db_version,10,child_id=0))
#     #     db.add(BoolConfig(ConfigEnum.db_version,11,child_id=0))
#     #     return
#     if hconfig(ConfigEnum.is_parent) and ParentDomain.query.count()==0:
#         external_ip=hiddify.get_ip(4)
#         db.session.add(ParentDomain(domain=f"{external_ip}.sslip.io",show_domains=[]))
# Events.db_init_event.subscribe(db_init_event)


def admin_prehook(flaskadmin, admin_bp):
    from .ParentDomainAdmin import ParentDomainAdmin
    flaskadmin.add_view(ParentDomainAdmin(ParentDomain, db.session))
    flaskadmin.add_view(ProxyDetailsAdmin(Proxy, db.session))
    # CommercialSettings.register(admin_bp)


Events.admin_prehook.subscribe(admin_prehook)


# def db_prehook():
# flaskadmin.add_view(ParentDomainAdmin(ParentDomain, db.session))
# from . import ProxyDominAdmin
# CommercialSettings.register(admin_bp)

# Events.db_prehook.subscribe(db_prehook)



import flask_bootstrap
import hiddifypanel
from dynaconf import FlaskDynaconf
from flask import Flask, request, g
from flask_babelex import Babel
from hiddifypanel.panel.init_db import init_db
import hiddifypanel
from hiddifypanel.models import *
from dotenv import dotenv_values
import os
from hiddifypanel.panel import hiddify


def create_app(cli=False, **config):
    app = Flask(__name__, static_url_path="/<proxy_path>/static/", instance_relative_config=True)

    for c, v in dotenv_values('app.cfg').items():
        if v.isdecimal():
            v = int(v)
        else:
            v = True if v.lower() == "true" else (False if v.lower() == "false" else v)

        app.config[c] = v

    app.jinja_env.line_statement_prefix = '%'
    app.jinja_env.filters['b64encode'] = hiddify.do_base_64

    app.is_cli = cli
    flask_bootstrap.Bootstrap4(app)

    hiddifypanel.panel.database.init_app(app)
    with app.app_context():
        init_db()

    hiddifypanel.panel.common.init_app(app)
    hiddifypanel.panel.admin.init_app(app)
    hiddifypanel.panel.user.init_app(app)
    hiddifypanel.panel.commercial.init_app(app)
    hiddifypanel.panel.cli.init_app(app)

    app.config.update(config)  # Override with passed config
    app.config['WTF_CSRF_CHECK_DEFAULT'] = False

    babel = Babel(app)

    @babel.localeselector
    def get_locale():
        # Put your logic here. Application can store locale in
        # user profile, cookie, session, etc.
        from hiddifypanel.models import ConfigEnum, hconfig
        if "admin" in request.base_url:
            g.locale = hconfig(ConfigEnum.admin_lang) or hconfig(ConfigEnum.lang) or 'fa'
        else:
            g.locale = hconfig(ConfigEnum.lang) or "fa"
        return g.locale

    from flask_wtf.csrf import CSRFProtect

    csrf = CSRFProtect(app)

    @app.before_request
    def check_csrf():
        if "/admin/user/" in request.base_url:
            return
        if "/admin/domain/" in request.base_url:
            return
        if "/admin/actions/" in request.base_url:
            return
        if "/api/v1/" in request.base_url:
            return
        csrf.protect()

    @app.after_request
    def apply_no_robot(response):
        response.headers["X-Robots-Tag"] = "noindex, nofollow";
        return response
    app.jinja_env.globals['get_locale'] = get_locale
    app.jinja_env.globals['version'] = hiddifypanel.__version__
    app.jinja_env.globals['static_url_for'] = hiddify.static_url_for

    return app


def create_app_wsgi():
    # workaround for Flask issue
    # that doesn't allow **config
    # to be passed to create_app
    # https://github.com/pallets/flask/issues/4170
    app = create_app()
    return app


# def create_cli_app():
#     # workaround for Flask issue
#     # that doesn't allow **config
#     # to be passed to create_app
#     # https://github.com/pallets/flask/issues/4170
#     app = create_app(cli=True)
#     return app



#!/usr/bin/env python3
import pprint
from flask_babelex import gettext as _
import pathlib
from hiddifypanel.models import *

from datetime import datetime, timedelta, date
import os
import sys
import json
import urllib.request
import subprocess
import re
from hiddifypanel.panel import hiddify, usage
from flask import current_app, render_template, request, Response, Markup, url_for, make_response, redirect
from hiddifypanel.panel.hiddify import flash

from flask_classful import FlaskView, route


class Actions(FlaskView):

    @hiddify.super_admin
    def index(self):
        return render_template('index.html')

    @hiddify.super_admin
    def reverselog(self, logfile):
        if logfile == None:
            return self.viewlogs()
        config_dir = current_app.config['HIDDIFY_CONFIG_PATH']

        with open(f'{config_dir}/log/system/{logfile}') as f:
            lines = [line for line in f]
            # logs="".join(lines[::-1])
            logs = "".join(lines)
        # resp= Response()
        # resp.mimetype="text/plain"
        from ansi2html import Ansi2HTMLConverter
        conv = Ansi2HTMLConverter()

        out = f'<div style="background-color:black; color:white;padding:10px">{conv.convert(logs)}</div>'
        # if len(lines)>5 and "----Finished!---" in "".join(lines[-min(10,len(lines)):]):
        #     out=f"<a href='#' target='_blank'><div style='background-color:#b1eab1; padding: 10px;border: solid;'>Finished! For scrolling the log click here.</div></a>{out}"

        response = make_response(out)
        response.headers['Access-Control-Allow-Origin'] = '*'
        response.headers['Access-Control-Allow-Methods'] = 'GET, POST, PUT, DELETE'
        response.headers['Access-Control-Allow-Headers'] = 'Content-Type'
        return response

    @hiddify.super_admin
    def viewlogs(self):
        config_dir = current_app.config['HIDDIFY_CONFIG_PATH']
        res = []
        for filename in sorted(os.listdir(f'{config_dir}/log/system/')):
            res.append(f"<a href='{url_for('admin.Actions:reverselog',logfile=filename)}'>{filename}</a>")
        return Markup("<br>".join(res))

    @route('apply_configs', methods=['POST'])
    def apply_configs(self):
        return self.reinstall(False)

    @route('reset', methods=['POST'])
    def reset(self):
        return self.reset2()

    @hiddify.super_admin
    def reset2(self):
        status = self.status()
        # flash(_("rebooting system may takes time please wait"),'info')
        # os.system(f"echo 'reboot'|at now + 3s ")

        resp = render_template("result.html",
                               out_type="info",
                               out_msg="",
                               log_path=get_logpath("restart.log"),
                               show_success=True,
                               domains=get_domains(),



                               )
        file = "restart.sh"
        config = current_app.config
        subprocess.Popen(f"sudo {config['HIDDIFY_CONFIG_PATH']}/{file}".split(" "), cwd=f"{config['HIDDIFY_CONFIG_PATH']}", start_new_session=True)
        import time
        time.sleep(1)
        return resp

    @route('reinstall', methods=['POST'])
    def reinstall(self, complete_install=True, domain_changed=False):
        return self.reinstall2(complete_install, domain_changed)

    @hiddify.super_admin
    def reinstall2(self, complete_install=True, domain_changed=False):
        if int(hconfig(ConfigEnum.db_version)) < 9:
            return ("Please update your panel before this action.")
        # if hconfig(ConfigEnum.parent_panel):
        #     try:
        #         hiddify_api.sync_child_to_parent()
        #     except e as Exception:
        #         flash(_('can not sync child with parent panel')+" "+e)

        domain_changed = request.args.get("domain_changed", str(domain_changed)).lower() == "true"
        complete_install = request.args.get("complete_install", str(complete_install)).lower() == "true"
        if domain_changed:
            flash((_('Your domains changed. Please do not forget to copy admin links, otherwise you can not access to the panel anymore.')), 'info')
        # flash(f'complete_install={complete_install} domain_changed={domain_changed} ', 'info')
        # return render_template("result.html")
        config = current_app.config
        hiddify.add_temporary_access()
        file = "install.sh" if complete_install else "apply_configs.sh"
        try:
            server_ip = urllib.request.urlopen('https://v4.ident.me/').read().decode('utf8')
        except:
            server_ip = "server_ip"

        # subprocess.Popen(f"{config_dir}/update.sh",env=my_env,cwd=f"{config_dir}")
        # os.system(f'cd {config_dir};{env} ./install.sh &')
        # rc = subprocess.call(f"cd {config_dir};./{file} & disown",shell=True)

        admin_secret = get_super_admin_secret()
        proxy_path = hconfig(ConfigEnum.proxy_path)
        admin_links = f"<h5 >{_('Admin Links')}</h5><ul>"
        admin_links += f"<li><span class='badge badge-danger'>{_('Not Secure')}</span>: <a class='badge ltr share-link' href='http://{server_ip}/{proxy_path}/{admin_secret}/admin/'>http://{server_ip}/{proxy_path}/{admin_secret}/admin/</a></li>"
        domains = get_panel_domains()
        # domains=[*domains,f'{server_ip}.sslip.io']
        for d in domains:
            link = f'https://{d}/{proxy_path}/{admin_secret}/admin/'
            admin_links += f"<li><a target='_blank' class='badge ltr' href='{link}'>{link}</a></li>"

        resp = render_template("result.html",
                               out_type="info",
                               out_msg=_("Success! Please wait around 4 minutes to make sure everything is updated. During this time, please save your proxy links which are:") +
                               admin_links,
                               log_path=get_logpath("0-install.log"),
                               show_success=True,
                               domains=get_domains(),



                               )

        subprocess.Popen(f"sudo {config['HIDDIFY_CONFIG_PATH']}/{file}".split(" "), cwd=f"{config['HIDDIFY_CONFIG_PATH']}", start_new_session=True)
        import time
        time.sleep(1)
        return resp

    @hiddify.super_admin
    def change_reality_keys(self):
        key = hiddify.generate_x25519_keys()
        set_hconfig(ConfigEnum.reality_private_key, key['private_key'])
        set_hconfig(ConfigEnum.reality_public_key, key['public_key'])
        hiddify.flash_config_success(restart_mode='apply', domain_changed=False)
        return redirect(url_for('admin.SettingAdmin:index'))

    @hiddify.super_admin
    def status(self):
        # hiddify.add_temporary_access()
        config = current_app.config
        # configs=read_configs()
        # subprocess.Popen(f"{config_dir}/update.sh",env=my_env,cwd=f"{config_dir}")
        # os.system(f'cd {config_dir};{env} ./install.sh &')
        # rc = subprocess.call(f"cd {config_dir};./{file} & disown",shell=True)
        from urllib.parse import urlparse

        o = urlparse(request.base_url)
        domain = o.hostname
        subprocess.Popen(f"sudo {config['HIDDIFY_CONFIG_PATH']}/status.sh".split(" "), cwd=f"{config['HIDDIFY_CONFIG_PATH']}", start_new_session=True)

        return render_template("result.html",
                               out_type="info",
                               out_msg=_("see the log in the bellow screen"),
                               log_path=get_logpath(f"status.log"),
                               # log_path=f"status.log",
                               show_success=False,
                               domains=get_domains()
                               )

    # @hiddify.super_admin

    @route('update', methods=['POST'])
    def update(self):
        return self.update2()

    @hiddify.super_admin
    def update2(self):
        hiddify.add_temporary_access()
        config = current_app.config
        cwd = os.getcwd()

        # os.chdir(config_dir)
        # rc = subprocess.call(f"./install.sh &",shell=True)
        # rc = subprocess.call(f"cd {config_dir};./update.sh & disown",shell=True)
        # os.system(f'cd {config_dir};./update.sh &')

        subprocess.Popen(f"sudo {config['HIDDIFY_CONFIG_PATH']}/update.sh".split(" "), cwd=f"{config['HIDDIFY_CONFIG_PATH']}", start_new_session=True)
        return render_template("result.html",
                               out_type="success",
                               out_msg=_("Success! Please wait around 5 minutes to make sure everything is updated."),
                               show_success=True,
                               log_path=get_logpath(f"update.log"),
                               domains=get_domains(),
                               )

    def get_some_random_reality_friendly_domain(self):
        test_domain = request.args.get("test_domain")
        import ping3
        from hiddifypanel.panel.clean_ip import ipasn, ipcountry
        ipv4 = hiddify.get_ip(4)
        server_country = (ipcountry.get(ipv4) or {}).get('country', {}).get('iso_code', 'unknown')
        server_asn = (ipasn.get(ipv4) or {}).get('autonomous_system_organization', 'unknown')
        res = "<table><tr><th>Domain</th><th>IP</th><th>Country</th><th>ASN</th><th>Ping (ms)</th><th>TCP ping (ms)</th></tr>"
        res += f"<tr><td>Your Server</td><td>{ipv4}</td><td>{server_country}</td><td>{server_asn}</td><td>0</td></tr>"
        import time
        start = time.time()
        for d in [test_domain, *hiddify.get_random_domains(30)]:
            if not d:
                continue
            if time.time()-start > 20:
                break
            print(d)
            tcp_ping = hiddify.is_domain_reality_friendly(d)
            if tcp_ping:
                dip = hiddify.get_domain_ip(d)
                dip_country = (ipcountry.get(dip) or {}).get('country', {}).get('iso_code', 'unknown')
                if dip_country == "IR":
                    continue
                response_time = -1
                try:
                    response_time = ping3.ping(d, unit='ms')
                    if response_time:
                        response_time = int(response_time)
                except:
                    pass
                dip_asn = (ipasn.get(dip) or {}).get('autonomous_system_organization', 'unknown')
                res += f"<tr><td>{d}</td><td>{dip}</td><td>{dip_country}</td><td>{dip_asn}</td><td>{response_time}</td><td>{tcp_ping}<td></tr>"

        return res+"</table>"

    @hiddify.super_admin
    def update_usage(self):

        import json

        return render_template("result.html",
                               out_type="info",
                               out_msg=f'<pre class="ltr">{json.dumps(usage.update_local_usage(),indent=2)}</pre>',
                               log_path=None
                               )


def get_logpath(logfile):
    return f'{hiddify.get_admin_path()}actions/reverselog/{logfile}/'


def get_domains():
    return [d.domain.replace("*", hiddify.get_random_string(3, 6)) for d in get_panel_domains(always_add_all_domains=True, always_add_ip=True)]



from hiddifypanel.cache import cache
from sqlalchemy_serializer import SerializerMixin
from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from flask_admin import Admin
from flask_admin.contrib.sqla import ModelView
from wtforms import SelectMultipleField

from sqlalchemy.orm import backref

from flask import Markup
from dateutil import relativedelta
from hiddifypanel.panel.database import db
import enum
import datetime
import uuid as uuid_mod
from enum import auto
from strenum import StrEnum
from hiddifypanel import Events

from .config_enum import ConfigEnum, ConfigCategory


class BoolConfig(db.Model, SerializerMixin):
    child_id = db.Column(db.Integer, db.ForeignKey('child.id'), primary_key=True, default=0)
    # category = db.Column(db.String(128), primary_key=True)
    key = db.Column(db.Enum(ConfigEnum), primary_key=True)
    value = db.Column(db.Boolean)

    def to_dict(d):
        return {
            'key': d.key,
            'value': d.value,
            'child_unique_id': d.child.unique_id if d.child else ''
        }


class StrConfig(db.Model, SerializerMixin):
    child_id = db.Column(db.Integer, db.ForeignKey('child.id'), primary_key=True, default=0)
    # category = db.Column(db.String(128), primary_key=True)
    key = db.Column(db.Enum(ConfigEnum), primary_key=True, default=ConfigEnum.admin_secret)
    value = db.Column(db.String(2048))

    def to_dict(d):
        return {
            'key': d.key,
            'value': d.value,
            'child_unique_id': d.child.unique_id if d.child else ''
        }


@cache.cache(ttl=500)
def hconfig(key: ConfigEnum, child_id=0):
    value = None
    try:
        str_conf = StrConfig.query.filter(StrConfig.key == key, StrConfig.child_id == child_id).first()
        if str_conf:
            value = str_conf.value
        else:
            bool_conf = BoolConfig.query.filter(BoolConfig.key == key, BoolConfig.child_id == child_id).first()
            if bool_conf:
                value = bool_conf.value
            else:
                # if key == ConfigEnum.ssfaketls_fakedomain:
                #     return hdomain(DomainType.ss_faketls)
                # if key == ConfigEnum.telegram_fakedomain:
                #     return hdomain(DomainType.telegram_faketls)
                # if key == ConfigEnum.fake_cdn_domain:
                #     return hdomain(DomainType.fake_cdn)
                print(f'{key} not found ')
    except:
        print(f'{key} error!')
        raise

    return value


def set_hconfig(key: ConfigEnum, value, child_id=0, commit=True):
    # hconfig.invalidate(key, child_id)
    # get_hconfigs.invalidate(child_id)
    hconfig.invalidate(key, child_id)
    hconfig.invalidate(key, child_id=child_id)
    hconfig.invalidate(key=key, child_id=child_id)
    if child_id == 0:
        hconfig.invalidate(key)
    # hconfig.invalidate_all()
    get_hconfigs.invalidate_all()
    if key.type() == bool:
        dbconf = BoolConfig.query.filter(BoolConfig.key == key, BoolConfig.child_id == child_id).first()
        if not dbconf:
            dbconf = BoolConfig(key=key, value=value, child_id=child_id)
            db.session.add(dbconf)
    else:
        dbconf = StrConfig.query.filter(StrConfig.key == key, StrConfig.child_id == child_id).first()
        if not dbconf:
            dbconf = StrConfig(key=key, value=value, child_id=child_id)
            db.session.add(dbconf)
    old_v = dbconf.value
    dbconf.value = value
    Events.config_changed.notify(conf=dbconf, old_value=old_v)
    if commit:
        db.session.commit()


@cache.cache(ttl=500)
def get_hconfigs(child_id=0):
    return {**{u.key: u.value for u in BoolConfig.query.filter(BoolConfig.child_id == child_id).all()},
            **{u.key: u.value for u in StrConfig.query.filter(StrConfig.child_id == child_id).all()},
            # ConfigEnum.telegram_fakedomain:hdomain(DomainType.telegram_faketls),
            # ConfigEnum.ssfaketls_fakedomain:hdomain(DomainType.ss_faketls),
            # ConfigEnum.fake_cdn_domain:hdomain(DomainType.fake_cdn)
            }


def add_or_update_config(commit=True, child_id=0, override_unique_id=True, **config):
    print(config)
    c = config['key']
    ckey = ConfigEnum(c)
    if c == ConfigEnum.unique_id and not override_unique_id:
        return

    v = str(config['value']).lower() == "true" if ckey.type() == bool else config['value']
    if ckey in [ConfigEnum.db_version]:
        return
    set_hconfig(ckey, v, child_id, commit=commit)


def bulk_register_configs(hconfigs, commit=True, override_child_id=None, override_unique_id=True):
    from hiddifypanel.panel import hiddify
    for conf in hconfigs:
        # print(conf)
        if conf['key'] == ConfigEnum.unique_id and not override_unique_id:
            continue
        child_id = override_child_id if override_child_id is not None else hiddify.get_child(conf.get('child_unique_id', None))
        # print(conf)
        add_or_update_config(commit=False, child_id=child_id, **conf)
    if commit:
        db.session.commit()



from flask import g, jsonify, abort, render_template, request, send_from_directory
from hiddifypanel.models import *
from hiddifypanel.panel import hiddify
import uuid
from flask import g, send_from_directory, url_for, session, Markup
import traceback
import user_agents
import hiddifypanel
import datetime
from flask_babelex import gettext as _


def init_app(app):
    app.jinja_env.globals['ConfigEnum'] = ConfigEnum
    app.jinja_env.globals['DomainType'] = DomainType
    app.jinja_env.globals['UserMode'] = UserMode
    app.jinja_env.globals['hconfig'] = hconfig

    @app.errorhandler(Exception)
    def internal_server_error(e):
        if not hasattr(e, 'code') or e.code == 500:
            trace = traceback.format_exc()

            last_version = hiddify.get_latest_release_version('hiddifypanel')
            has_update = "dev" not in hiddifypanel.__version__ and f'{last_version}' != hiddifypanel.__version__
            return render_template('500.html', error=e, trace=trace, has_update=has_update, last_version=last_version), 500
        # if e.code in [400,401,403]:
        #     return render_template('access-denied.html',error=e), e.code

        return render_template('error.html', error=e), e.code

    @app.url_defaults
    def add_proxy_path_user(endpoint, values):

        if 'user_secret' not in values and hasattr(g, 'user_uuid'):
            values['user_secret'] = f'{g.user_uuid}'
        if 'proxy_path' not in values:
            # values['proxy_path']=f'{g.proxy_path}'
            values['proxy_path'] = hconfig(ConfigEnum.proxy_path)

    @app.route("/<proxy_path>/videos/<file>")
    def videos(file):
        print("file", file, app.config['HIDDIFY_CONFIG_PATH'] +
              '/hiddify-panel/videos/'+file)
        return send_from_directory(app.config['HIDDIFY_CONFIG_PATH']+'/hiddify-panel/videos/', file)
    # @app.template_filter()
    # def rel_datetime(value):
    #     diff=datetime.datetime.now()-value
    #     return format_timedelta(diff, add_direction=True, locale=hconfig(ConfigEnum.lang))

    @app.url_value_preprocessor
    def pull_secret_code(endpoint, values):
        # print("Y",endpoint, values)
        # if values is None:
        #     return
        # if hiddifypanel.__release_date__ + datetime.timedelta(days=40) < datetime.datetime.now() or hiddifypanel.__release_date__ > datetime.datetime.now():
        #     abort(400, _('This version of hiddify panel is outdated. Please update it from admin area.'))
        g.user = None
        g.user_uuid = None
        g.is_admin = False

        if request.args.get('darkmode') != None:
            session['darkmode'] = request.args.get('darkmode', '').lower() == 'true'
        g.darkmode = session.get('darkmode', False)
        import random
        g.install_pwa = random.random() <= 0.05
        if request.args.get('pwa') != None:
            session['pwa'] = request.args.get('pwa', '').lower() == 'true'
        g.pwa = session.get('pwa', False)

        g.user_agent = user_agents.parse(request.user_agent.string)
        if g.user_agent.is_bot:
            abort(400, "invalid")
        g.proxy_path = values.pop('proxy_path', None) if values else None
        if g.proxy_path != hconfig(ConfigEnum.proxy_path):
            if app.config['DEBUG']:
                abort(400, Markup(f"Invalid Proxy Path <a href=/{hconfig(ConfigEnum.proxy_path)}/{get_super_admin_secret()}/admin>admin</a>"))
            abort(400, "Invalid Proxy Path")
        if endpoint == 'static' or endpoint == "videos":
            return
        tmp_secret = values.pop('user_secret', None) if values else None
        try:
            if tmp_secret:
                g.user_uuid = uuid.UUID(tmp_secret)
        except:
            # raise PermissionError("Invalid secret")
            abort(400, 'invalid user')
        g.admin = get_admin_user_db(tmp_secret)
        g.is_admin = g.admin is not None
        bare_path = request.path.replace(g.proxy_path, "").replace(tmp_secret, "").lower()
        if not g.is_admin:
            g.user = User.query.filter(User.uuid == f'{g.user_uuid}').first()
            if not g.user:
                abort(401, 'invalid user')
            if endpoint and ("admin" in endpoint or "api" in endpoint):
                # raise PermissionError("Access Denied")
                abort(403, 'Access Denied')
            if "admin" in bare_path or "api" in bare_path:
                abort(403, 'Access Denied')

        if hconfig(ConfigEnum.telegram_bot_token):
            import hiddifypanel.panel.commercial.telegrambot as telegrambot
            if (not telegrambot.bot) or (not telegrambot.bot.username):
                telegrambot.register_bot()
            g.bot = telegrambot.bot
        # print(g.user)



from flask_admin.contrib import sqla
from hiddifypanel.panel.database import db
import datetime
from hiddifypanel.models import *
from flask import Markup, g, request, url_for, abort
from wtforms.validators import Regexp, ValidationError
import re
import uuid
from hiddifypanel.drivers import user_driver
from .adminlte import AdminLTEModelView
# from gettext import gettext as _
from flask_babelex import gettext as __
from flask_babelex import lazy_gettext as _

from wtforms.validators import NumberRange


from hiddifypanel.panel import hiddify, custom_widgets
from hiddifypanel.panel.hiddify import flash
from wtforms.fields import StringField
from flask_bootstrap import SwitchField


class UserAdmin(AdminLTEModelView):
    column_sortable_list = ["name", "current_usage_GB", 'mode', "remaining_days", "comment", 'last_online', "uuid", 'remaining_days']
    column_searchable_list = [("uuid"), "name"]
    column_list = ["name", "UserLinks", "current_usage_GB", "remaining_days", "comment", 'last_online', 'mode', 'admin', "uuid"]
    form_extra_fields = {
        'reset_days': SwitchField(_("Reset package days")),
        'reset_usage': SwitchField(_("Reset package usage")),
        # 'disable_user': SwitchField(_("Disable User"))
    }
    list_template = 'model/user_list.html'
    form_excluded_columns = ['monthly', 'telegram_id', 'last_online', 'expiry_time', 'last_reset_time', 'current_usage_GB',
                             'start_date', 'added_by', 'admin', 'details', 'max_ips', 'ed25519_private_key', 'ed25519_public_key']
    page_size = 50
    # edit_modal=True
    # create_modal=True
    # column_display_pk = True
    # can_export = True
    # form_overrides = dict(monthly=SwitchField)
    form_overrides = {
        'start_date': custom_widgets.DaysLeftField,
        'mode': custom_widgets.EnumSelectField
    }

    # form_overrides = dict(expiry_time=custom_widgets.DaysLeftField,last_reset_time=custom_widgets.LastResetField)
    form_widget_args = {
        'current_usage_GB': {'min': '0'},
        'usage_limit_GB': {'min': '0'},

    }
    form_args = {
        'max_ips': {
            'validators': [NumberRange(min=3, max=10000)]
        },
        'mode': {'enum': UserMode},
        'uuid': {
            'validators': [Regexp(r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$', message=__("Should be a valid uuid"))]
            #     'label': 'First Name',
            #     'validators': [required()]
        }
        # ,
        # 'expiry_time':{
        # "":'%Y-%m-%d'
        # }
    }
    # column_labels={'uuid':_("user.UUID")}
    # column_filters=["usage_limit_GB","current_usage_GB",'admin','is_active']

    column_labels = {
        "Actions": _("actions"),
        "name": _("user.name"),
        "UserLinks": _("user.user_links"),
        "usage_limit_GB": _("user.usage_limit_GB"),
        "monthly": _("Reset every month"),
        "mode": _("Mode"),
        "admin": _("Added by"),
        "current_usage_GB": _("user.current_usage_GB"),
        "start_date": _("Start Date"),
        "remaining_days": _("user.expiry_time"),
        "last_reset_time": _("user.last_reset_time"),
        "uuid": _("user.UUID"),
        "comment": _("Note"),
        'last_online': _('Last Online'),
        "package_days": _('Package Days'),
        "max_ips": _('Max IPs'),
        "enable": _('Enable'),

    }
    # can_set_page_size=True

    def search_placeholder(self):
        return f"{_('search')} {_('user.UUID')} {_('user.name')}"
    # def get_column_name(self,field):
    #         return "x"
    #  return column_labels[field]
    column_descriptions = dict(
        # name=_'just for remembering',
        # usage_limit_GB="in GB",
        # current_usage_GB="in GB"
        comment=_("Add some text that is only visible to you."),
        mode=_("Define the user mode. Should the usage reset every month?"),
        last_reset_time=_("If monthly is enabled, the usage will be reset after 30 days from this date."),
        start_date=_("From when the user package will be started? Empty for start from first connection"),
        package_days=_("How many days this package should be available?")
    )
    # column_editable_list=["usage_limit_GB","current_usage_GB","expiry_time"]
    # form_extra_fields={
    #     'uuid': {'label_name':"D"}

    #     }

    # can_edit = False
    # def on_model_change(self, form, model, is_created):
    #     model.password = generate_password_hash(model.password)

    def _name_formatter(view, context, model, name):
        proxy_path = hconfig(ConfigEnum.proxy_path)
        # print("model.telegram_id",model.telegram_id)
        extra = ""
        if hconfig(ConfigEnum.telegram_bot_token) and model.telegram_id:
            extra = f'<button class="btn hbtn bg-h-blue btn-xs " onclick="show_send_message({model.id})" ><i class="fa-solid fa-paper-plane"></i></button> '

        link = ''
        if model.is_active:
            link = '<i class="fa-solid fa-circle-check text-success"></i> '
        elif len(model.ips):
            link = '<i class="fa-solid fa-users-slash text-danger" title="{_("Too many Connected IPs")}"></i>'
        else:
            link = '<i class="fa-solid fa-circle-xmark text-danger"></i> '

        link += f"<a target='_blank' href='/{proxy_path}/{model.uuid}/#{model.name}'>{model.name} <i class='fa-solid fa-arrow-up-right-from-square'></i></a>"
        return Markup(extra+link)

    def _ul_formatter(view, context, model, name):
        domains = get_panel_domains()
        return Markup(" ".join([hiddify.get_user_link(model.uuid, d, 'new', model.name) for d in domains]))

    def _uuid_formatter(view, context, model, name):
        return Markup(f"<span>{model.uuid}</span>")
    # def _usage_formatter(view, context, model, name):
    #     return round(model.current_usage_GB,3)

    def _usage_formatter(view, context, model, name):
        u = round(model.current_usage_GB, 3)
        t = round(model.usage_limit_GB, 3)
        rate = round(u*100/(t+0.000001))
        state = "danger" if u >= t else ('warning' if rate > 80 else 'success')
        color = "#ff7e7e" if u >= t else ('#ffc107' if rate > 80 else '#9ee150')
        return Markup(f"""
        <div class="progress progress-lg position-relative" style="min-width: 100px;">
          <div class="progress-bar progress-bar-striped" role="progressbar" style="width: {rate}%;background-color: {color};" aria-valuenow="{rate}" aria-valuemin="0" aria-valuemax="100"></div>
              <span class='badge position-absolute' style="left:auto;right:auto;width: 100%;font-size:1em">{u} {_('user.home.usage.from')} {t} GB</span>

        </div>
        """)

    def _expire_formatter(view, context, model, name):
        remaining = remaining_days(model)

        diff = datetime.timedelta(days=remaining)

        state = 'success' if diff.days > 7 else ('warning' if diff.days > 0 else 'danger')
        formated = hiddify.format_timedelta(diff)
        return Markup(f"<span class='badge badge-{state}'>{'*' if not model.start_date else ''} {formated} </span>")
        # return Markup(f"<span class='badge ltr badge-}'>{days}</span> "+_('days'))

    def _admin_formatter(view, context, model, name):
        return Markup(f"<a href='{url_for('flask.user.index_view',admin_id=model.added_by)}' class='btn btn-xs btn-default'>{model.admin.name}</a>")

    def _online_formatter(view, context, model, name):
        if not model.last_online:
            return Markup("-")
        diff = model.last_online-datetime.datetime.now()

        if diff.days < -1000:
            return Markup("-")
        if diff.total_seconds() > -60*2:
            return Markup(f"<span class='badge badge-success'>{_('Online')}</span>")
        state = "danger" if diff.days < -3 else ("success" if diff.days >= -1 else "warning")
        return Markup(f"<span class='badge badge-{state}'>{hiddify.format_timedelta(diff,granularity='min')}</span>")

        # return Markup(f"<span class='badge ltr badge-{'success' if days>7 else ('warning' if days>0 else 'danger') }'>{days}</span> "+_('days'))

    column_formatters = {
        'name': _name_formatter,
        'UserLinks': _ul_formatter,
        'uuid': _uuid_formatter,
        'current_usage_GB': _usage_formatter,
        "remaining_days": _expire_formatter,
        'last_online': _online_formatter,
        'admin': _admin_formatter
    }

    def on_model_delete(self, model):
        if len(User.query.all()) <= 1:
            raise ValidationError(f"at least one user should exist")
        user_driver.remove_client(model)
        # hiddify.flash_config_success()

    # def is_accessible(self):
    #     return g.is_admin

    def on_form_prefill(self, form, id=None):
        # print("================",form._obj.start_date)
        if id == None or form._obj is None or form._obj.start_date is None:
            msg = _("Package not started yet.")
            # form.reset['class']="d-none"
            delattr(form, 'reset_days')
            delattr(form, 'reset_usage')
            # delattr(form,'disable_user')
        else:
            remaining = remaining_days(form._obj)
            msg = _("Remaining: ") + hiddify.format_timedelta(datetime.timedelta(days=remaining))
            form.reset_days.label.text += f" ({msg})"
            usr_usage = f" ({_('user.home.usage.title')} {round(form._obj.current_usage_GB,3)}GB)"
            form.reset_usage.label.text += usr_usage
            form.usage_limit_GB.label.text += usr_usage

        # if form._obj.mode==UserMode.disable:
        #     delattr(form,'disable_user')
        # form.disable_user.data=form._obj.mode==UserMode.disable
        form.package_days.label.text += f" ({msg})"

    def get_edit_form(self):
        form = super().get_edit_form()
        # print(form.__dict__)
        # user=User.query.filter(User.uuid==form.uuid).first()
        # if user and user.start_date:
        #     form.reset = SwitchField("Reset")
        return form

    def on_model_change(self, form, model, is_created):
        model.max_ips = max(3, model.max_ips or 10000)
        if len(User.query.all()) % 4 == 0:
            flash(('<div id="show-modal-donation"></div>'), ' d-none')
        if not re.match("^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$", model.uuid):
            raise ValidationError('Invalid UUID e.g.,' + str(uuid.uuid4()))

        if form.reset_usage.data:
            model.current_usage_GB = 0
        # if form.disable_user.data:
        #     model.mode=UserMode.disable
        if form.reset_days.data:
            model.start_date = None
        model.package_days = min(model.package_days, 10000)
        old_user = user_by_id(model.id)
        if not model.added_by:
            model.added_by = g.admin.id
        if not g.admin.can_have_more_users():
            raise ValidationError(_('You have too much users! You can have only %(active)s active users and %(total)s users',
                                  active=g.admin.max_active_users, total=g.admin.max_users))
        if old_user and old_user.uuid != model.uuid:
            user_driver.remove_client(old_user)
        if not model.ed25519_private_key:
            priv, publ = hiddify.get_ed25519_private_public_pair()
            model.ed25519_private_key = priv
            model.ed25519_public_key = publ
        # model.expiry_time=datetime.date.today()+datetime.timedelta(days=model.expiry_time)

        # if model.current_usage_GB < model.usage_limit_GB:
        #     xray_api.add_client(model.uuid)
        # else:
        #     xray_api.remove_client(model.uuid)
        # hiddify.flash_config_success()
    # def is_accessible(self):
    #     return is_valid()

    def after_model_change(self, form, model, is_created):

        user = User.query.filter(User.uuid == model.uuid).first()
        if is_user_active(user):
            user_driver.add_client(model)
        else:
            user_driver.remove_client(model)
        hiddify.quick_apply_users()

    def after_model_delete(self, model):
        user_driver.remove_client(model)

        hiddify.quick_apply_users()

    def get_list(self, page, sort_column, sort_desc, search, filters, *args, **kwargs):
        res = None
        # print('aaa',args, kwargs)
        if 'remaining_days' == sort_column:
            query = self.get_query()

            data = query.all()
            count = len(data)
            data = sorted(data, key=lambda p: p.remaining_days, reverse=sort_desc)
            res = count, data
        else:

            res = super().get_list(page, sort_column, sort_desc, search, filters, *args, **kwargs)
        return res

        # Override the default get_list method to use the custom sort function
        query = self.session.query(self.model)
        if self._sortable_columns:
            # print("sor",self._sortable_columns['remaining_days'])
            for column, direction in self._get_default_order():
                # if column == 'remaining_days':
                #     # Use the custom sort function for 'remaining_days'
                #     query = query.order_by(self.model.remaining_days.asc() if direction == 'asc' else self.model.remaining_days.desc())
                # else:
                # Use the default sort function for other columns
                query = query.order_by(getattr(self.model, column).asc() if direction == 'asc' else getattr(self.model, column).desc())
        count = query.count()
        data = query.all()
        return count, data

        # Override get_query() to filter rows based on a specific condition

    def get_query(self):
        # Get the base query
        query = super().get_query()

        admin_id = int(request.args.get("admin_id") or g.admin.id)
        if admin_id not in g.admin.recursive_sub_admins_ids():
            abort(403)
        admin = AdminUser.query.filter(AdminUser.id == admin_id).first()
        if not admin:
            abort(403)

        query = query.filter(User.added_by.in_(admin.recursive_sub_admins_ids()))

        return query

    # Override get_count_query() to include the filter condition in the count query
    def get_count_query(self):
        # Get the base count query
        query = self.get_query()
        from sqlalchemy import func

        query = query.session.query(func.count(User.id))
        # query = super().get_count_query()
        # admin_id=int(request.args.get("admin_id") or g.admin.id)
        # if admin_id not in g.admin.recursive_sub_admins_ids():
        #     abort(403)
        # admin=AdminUser.query.filter(AdminUser.id==admin_id).first()
        # if not admin:
        #     abort(403)

        return query
